# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from typing import Optional

import pandas as pd


def relative_response_rates(
    df: pd.DataFrame,
    df_target: Optional[pd.DataFrame] = None,
    per_column: bool = True,
) -> pd.DataFrame:
    """Produces a summary table of number of responses and proportion of completed responses.

    Args:
        df (pd.DataFrame): A DataFrame to calculate aggregated response rates for.
        df_target (Optional[pd.DataFrame], optional): Defaults to None.
            Determines what is the denominator from which notnull are a fraction of.
            If None - it's the number of rows in df.
            If some df is provided - then it is assumed that df is a subset of df_target, and the
            response rate is calculated as the fraction of notnull values in df from (divided by)
            the number of notnull values in df_target.
        per_column (bool, optional): Default is True.
            The per_column argument is relevant only if df_target is other than None (i.e.: trying to compare df to some df_target).
            If per_column is True (default) - it indicates that the relative response rates of columns in df will be
                by comparing each column in df to the same column in target.
                If this is True, the columns in df and df_target must be identical.
            If per_column is False then df is compared to the overall number of nonnull rows in the target df.

    Returns:
        pd.DataFrame: A column per column in the original df, and two rows:
            One row with number of non-null observations, and
            A second row with the proportion of non-null observations.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots.general_stats import relative_response_rates

            df = pd.DataFrame({"o1": (7, 8, 9, 10), "o2": (7, 8, 9, np.nan), "id": (1, 2, 3, 4)})

            relative_response_rates(df).to_dict()

                # {'o1': {'n': 4.0, '%': 100.0},
                # 'o2': {'n': 3.0, '%': 75.0},
                # 'id': {'n': 4.0, '%': 100.0}}

            df_target = pd.concat([df, df])
            relative_response_rates(df, df_target).to_dict()

                # {'o1': {'n': 4.0, '%': 50.0},
                #  'o2': {'n': 3.0, '%': 50.0},
                #  'id': {'n': 4.0, '%': 50.0}}


            # Dividing by number of total notnull rows in df_rarget
            df_target.notnull().all(axis=1).sum()  # == 6
            relative_response_rates(df, df_target, False).to_dict()

                # {'o1': {'n': 4.0, '%': 66.66666666666666},
                # 'o2': {'n': 3.0, '%': 50.0},
                # 'id': {'n': 4.0, '%': 66.66666666666666}}

    """
    df_n_notnull_rows = df.notnull().sum()

    if df_target is None:
        target_n_notnull_rows = df.shape[0]
    elif per_column:  # number of notnull rows, *per column*, in df_target

        # verify that the columns of df and df_target are identical:
        if (len(df.columns) != len(df_target.columns)) or (
            df.columns.tolist() != df_target.columns.tolist()
        ):
            raise ValueError(
                f"""
                df and df_target must have the exact same columns.
                Instead, thes column names are, (df, df_target) = ({df.columns.tolist()}, {df_target.columns.tolist()})
                """
            )

        # If they are, we can proceed forward:
        target_n_notnull_rows = df_target.notnull().sum()
    else:  # number of notnull *rows* (i.e.: complete rows) in df_target
        target_n_notnull_rows = df_target.notnull().all(axis=1).sum()

    if any(df_n_notnull_rows > target_n_notnull_rows):
        raise ValueError(
            f"""
            The number of (notnull) rows in df MUST be smaller or equal to the number of rows in df_target.
            These were, (df_n_notnull_rows, target_n_notnull_rows) = ({df_n_notnull_rows}, {target_n_notnull_rows})
            """
        )

    return pd.DataFrame(
        {
            "n": df_n_notnull_rows,
            "%": 100 * (df_n_notnull_rows / target_n_notnull_rows),
        }
    ).transpose()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import balance.testutil
import pandas as pd

from balance.sample_class import Sample
from balance.weighting_methods import adjust_null as balance_adjust_null


sample = Sample.from_frame(
    df=pd.DataFrame(
        {
            "a": (1, 2, 3, 1),
            "b": (-42, 8, 2, -42),
            "o": (7, 8, 9, 10),
            "c": ("x", "y", "z", "x"),
            "id": (1, 2, 3, 4),
            "w": (0.5, 2, 1, 1),
        }
    ),
    id_column="id",
    weight_column="w",
    outcome_columns="o",
)

target = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3),
            "b": (-42, 8, 2),
            "c": ("x", "y", "z"),
            "id": (1, 2, 3),
            "w": (2, 0.5, 1),
        }
    ),
    id_column="id",
    weight_column="w",
)


class TestAdjustmentNull(
    balance.testutil.BalanceTestCase,
):
    def test_adjust_null(self):
        res = balance_adjust_null.adjust_null(
            pd.DataFrame({"a": [1, 2, 3]}),
            pd.Series([4, 5, 6]),
            pd.DataFrame({"a": [7, 8, 9]}),
            pd.Series([10, 11, 12]),
        )
        self.assertEqual(res["weight"], pd.Series([4, 5, 6]))
        self.assertEqual(res["model"]["method"], "null_adjustment")

        result = sample.adjust(target, method="null")
        self.assertEqual(sample.weights().df, result.weights().df)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd

from balance.stats_and_plots.weights_stats import _check_weights_are_valid
from balance.util import model_matrix, rm_mutual_nas
from scipy.stats import norm

from statsmodels.stats.weightstats import DescrStatsW

logger: logging.Logger = logging.getLogger(__package__)


##########################################
# Weighted statistics
##########################################


def _prepare_weighted_stat_args(
    v: Union[  # pyre-ignore[11]: np.matrix is a type
        List,
        pd.Series,
        pd.DataFrame,
        np.ndarray,
        np.matrix,
    ],
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    inf_rm: bool = False,
) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Checks that the values (v) and weights (w) are of the supported types and of the same length. Can also replace np.Inf to np.nan.

    Args:
        v (Union[List, pd.Series, pd.DataFrame, np.ndarray, np.matrix,]): a series of values. Notice that pd.DataFrame and np.matrix should store the Series/np.ndarry as columns (not rows).
        w (Union[List, pd.Series, np.ndarray, None]): a series of weights to be used with v (could also be none). Defaults to None.
        inf_rm (bool, optional): should np.inf values be replaced by np.nan? Defaults to False.

    Returns:
        Tuple[pd.DataFrame, pd.Series]: the original v and w after they have been validated, and turned to pd.DataFrame and pd.Series (if needed).
        The values might be int64 or float64, depending on the input.
    """

    supported_types = (
        list,
        pd.Series,
        np.ndarray,
    )

    tmp_supported_types = supported_types + (
        pd.DataFrame,
        np.matrix,
    )
    if type(v) not in tmp_supported_types:
        raise TypeError(f"argument must be {tmp_supported_types}, is {type(v)}")

    tmp_supported_types = supported_types + (type(None),)
    if type(w) not in supported_types + (type(None),):
        raise TypeError(f"argument must be {tmp_supported_types}, is {type(w)}")

    # NOTE: np.matrix is an instance of np.ndarray, so we must turn it to pd.Dataframe before moving forward.
    if isinstance(v, np.matrix):
        v = pd.DataFrame(v)
    if isinstance(v, np.ndarray) or isinstance(v, list):
        v = pd.Series(v)
    if isinstance(v, pd.Series):
        v = v.to_frame()
    if isinstance(w, np.ndarray) or isinstance(w, list):
        w = pd.Series(w)
    if w is None:
        w = pd.Series(np.ones(len(v)), index=v.index)

    if v.shape[0] != w.shape[0]:
        raise ValueError(
            f"weights (w) and data (v) must have same number of rows, ({v.shape[0]}, {w.shape[0]})"
        )

    dtypes = v.dtypes if hasattr(v.dtypes, "__iter__") else [v.dtypes]

    if not all(np.issubdtype(x, np.number) for x in dtypes):
        raise TypeError("all columns must be numeric")

    if inf_rm:
        v = v.replace([np.inf, -np.inf], np.nan)
        w = w.replace([np.inf, -np.inf], np.nan)
    v = v.reset_index(drop=True)
    w = w.reset_index(  # pyre-ignore[16]: w is a pd.Series which has a reset_index method.
        drop=True
    )

    _check_weights_are_valid(w)

    return v, w


def weighted_mean(
    v: Union[
        List,
        pd.Series,
        pd.DataFrame,
        np.matrix,
    ],
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    inf_rm: bool = False,
) -> pd.Series:
    """
    Computes the weighted average of a pandas Series or DataFrame.

    If no weights are supplied, it just computes the simple arithmetic mean.

    See:
    https://en.wikipedia.org/wiki/Weighted_arithmetic_mean

    Uses :func:`_prepare_weighted_stat_args`.

    Args:
        v (Union[ List, pd.Series, pd.DataFrame, np.matrix, ]): values to average. None (or np.nan) values are treated like 0.
            If v is a DataFrame than the average of the values from each column will be returned, all using the same set of weights from w.
        w (Union[ List, pd.Series], optional): weights. Defaults to None.
            If there is None value in the weights, that value will be ignored from the calculation.
        inf_rm (bool, optional): whether to remove infinite (from weights or values) from the computation. Defaults to False.

    Returns:
        pd.Series(dtype=np.float64): The weighted mean.
        If v is a DataFrame with several columns than the pd.Series will have a value for the weighted mean of each of the columns.
        If inf_rm=False then:
            If v has Inf then the output will be Inf.
            If w has Inf then the output will be np.nan.

    Examples:
        ::
            from balance.stats_and_plots.weighted_stats import weighted_mean

            weighted_mean(pd.Series((1, 2, 3, 4)))
                # 0    2.5
                # dtype: float64

            weighted_mean(pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4)))
                # 0    3.0
                # dtype: float64

            df = pd.DataFrame(
                {"a": [1,2,3,4], "b": [1,1,1,1]}
            )
            w = pd.Series((1, 2, 3, 4))
            weighted_mean(df, w)
                # a    3.0
                # b    1.0
                # dtype: float64
    """
    v, w = _prepare_weighted_stat_args(v, w, inf_rm)
    return v.multiply(w, axis="index").sum() / w.sum()


def var_of_weighted_mean(
    v: Union[List, pd.Series, pd.DataFrame, np.matrix],
    w: Optional[Union[List, pd.Series, np.ndarray]] = None,
    inf_rm: bool = False,
) -> pd.Series:
    """
    Computes the variance of the weighted average (pi estimator for ratio-mean) of a list of values and their corresponding weights.

    If no weights are supplied, it assumes that all values have equal weights of 1.0.

    See: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Variance_of_the_weighted_mean_(%CF%80-estimator_for_ratio-mean)

    Uses :func:`_prepare_weighted_stat_args`.

    v (Union[List, pd.Series, pd.DataFrame, np.matrix]): A series of values. If v is a DataFrame, the weighted variance will be calculated for each column using the same set of weights from `w`.
    w (Optional[Union[List, pd.Series, np.ndarray]]): A series of weights to be used with `v`. If None, all values will be weighted equally.
    inf_rm (bool, optional): Whether to remove infinite (from weights or values) from the computation. Defaults to False.

    Returns:
        pd.Series: The variance of the weighted mean. If `v` is a DataFrame with several columns, the pd.Series will have a value for the weighted variance of each column. The values are of data type `np.float64`.
        If `inf_rm` is False:
            If `v` has infinite values, the output will be `Inf`.
            If `w` has infinite values, the output will be `np.nan`.


    Examples:
        ::
            from balance.stats_and_plots.weighted_stats import var_of_weighted_mean

            #  In R: sum((1:4 - mean(1:4))^2 / 4) / (4)
            #  [1] 0.3125
            var_of_weighted_mean(pd.Series((1, 2, 3, 4)))
                # pd.Series(0.3125)

            #  For a reproducible R example, see: https://gist.github.com/talgalili/b92cd8cdcbfc287e331a8f27db265c00
            var_of_weighted_mean(pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4)))
                # pd.Series(0.24)

            df = pd.DataFrame(
                {"a": [1,2,3,4], "b": [1,1,1,1]}
            )
            w = pd.Series((1, 2, 3, 4))
            var_of_weighted_mean(df, w)

                # a    0.24
                # b    0.00
                # dtype: float64
    """
    v, w = _prepare_weighted_stat_args(v, w, inf_rm)
    weighed_mean_of_v = weighted_mean(v, w, inf_rm)  # This is a pd.Series
    # NOTE: the multiply needs to be called from the pd.Series. Unfortunately this makes the code less readable.
    sum_of_squared_weighted_diffs = (
        ((v - weighed_mean_of_v).multiply(w, axis="index")).pow(2).sum()
    )
    squared_sum_of_w = w.sum() ** 2  # This is a pd.series
    return sum_of_squared_weighted_diffs / squared_sum_of_w


def ci_of_weighted_mean(
    v: Union[List[float], pd.Series, pd.DataFrame, np.ndarray],
    w: Optional[Union[List[float], pd.Series, np.ndarray]] = None,
    conf_level: float = 0.95,
    round_ndigits: Optional[int] = None,
    inf_rm: bool = False,
) -> pd.Series:
    """
    Computes the confidence interval of the weighted mean of a list of values and their corresponding weights.

    If no weights are supplied, it assumes that all values have equal weights of 1.0.

    v (Union[List[float], pd.Series, pd.DataFrame, np.ndarray]): A series of values. If v is a DataFrame, the weighted mean and its confidence interval will be calculated for each column using the same set of weights from `w`.
    w (Optional[Union[List[float], pd.Series, np.ndarray]]): A series of weights to be used with `v`. If None, all values will be weighted equally.
    conf_level (float, optional): Confidence level for the interval, between 0 and 1. Defaults to 0.95.
    round_ndigits (Optional[int], optional): Number of decimal places to round the confidence interval. If None, the values will not be rounded. Defaults to None.
    inf_rm (bool, optional): Whether to remove infinite (from weights or values) from the computation. Defaults to False.

    Returns:
        pd.Series: The confidence interval of the weighted mean. If `v` is a DataFrame with several columns, the pd.Series will have a value for the confidence interval of each column. The values are of data type Tuple[np.float64, np.float64].
        If `inf_rm` is False:
            If `v` has infinite values, the output will be `Inf`.
            If `w` has infinite values, the output will be `np.nan`.

    Examples:
        ::
            from balance.stats_and_plots.weighted_stats import ci_of_weighted_mean

            ci_of_weighted_mean(pd.Series((1, 2, 3, 4)))
                # 0    (1.404346824279273, 3.5956531757207273)
                # dtype: object

            ci_of_weighted_mean(pd.Series((1, 2, 3, 4)), round_ndigits = 3).to_list()
                # [(1.404, 3.596)]

            ci_of_weighted_mean(pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4)))
                # 0    (2.039817664728938, 3.960182335271062)
                # dtype: object

            ci_of_weighted_mean(pd.Series((1, 2, 3, 4)), pd.Series((1, 2, 3, 4)), round_ndigits = 3).to_list()
                # [(2.04, 3.96)]

            df = pd.DataFrame(
                {"a": [1,2,3,4], "b": [1,1,1,1]}
            )
            w = pd.Series((1, 2, 3, 4))
            ci_of_weighted_mean(df, w, conf_level = 0.99, round_ndigits = 3)
                # a    (1.738, 4.262)
                # b        (1.0, 1.0)
                # dtype: object
            ci_of_weighted_mean(df, w, conf_level = 0.99, round_ndigits = 3).to_dict()
                # {'a': (1.738, 4.262), 'b': (1.0, 1.0)}
    """
    v, w = _prepare_weighted_stat_args(v, w, inf_rm)
    weighed_mean_of_v = weighted_mean(v, w, inf_rm)
    var_weighed_mean_of_v = var_of_weighted_mean(v, w, inf_rm)
    z_value = norm.ppf((1 + conf_level) / 2)

    if isinstance(v, pd.Series):
        ci_index = v.index
    elif isinstance(v, pd.DataFrame):
        ci_index = v.columns
    else:
        ci_index = None

    ci = pd.Series(
        [
            (
                weighed_mean_of_v[i] - z_value * np.sqrt(var_weighed_mean_of_v[i]),
                weighed_mean_of_v[i] + z_value * np.sqrt(var_weighed_mean_of_v[i]),
            )
            for i in range(len(weighed_mean_of_v))
        ],
        index=ci_index,
    )

    if round_ndigits is not None:
        # Apply a lambda function to round a pd.Series of tuples to x decimal places
        ci = ci.apply(lambda t: tuple(round(x, round_ndigits) for x in t))

    # pyre-ignore[7]: pyre thinks this function could return a DataFrame because of ci = ci.apply(round_tuple). It's wrong.
    return ci


def weighted_var(
    v: Union[
        List,
        pd.Series,
        pd.DataFrame,
        np.ndarray,
        np.matrix,
    ],
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    inf_rm: bool = False,
) -> pd.Series:
    """
    Calculate the sample weighted variance (a.k.a 'reliability weights').
    This is described here:
    https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Reliability_weights_2
    And also used in SDMTools, see:
    https://www.gnu.org/software/gsl/doc/html/statistics.html#weighted-samples

    Uses :func:`weighted_mean` and :func:`_prepare_weighted_stat_args`.

    Args:
        v (Union[ List, pd.Series, pd.DataFrame, np.matrix, ]): values to get the weighted variance for. None values are treated like 0.
            If v is a DataFrame than the average of the values from each column will be returned, all using the same set of weights from w.
        w (Union[ List, pd.Series], optional): weights. Defaults to None.
            If there is None value in the weights, that value will be ignored from the calculation.
        inf_rm (bool, optional): whether to remove infinite (from weights or values) from the computation. Defaults to False.

    Returns:
        pd.Series[np.float64]: The weighted variance.
        If v is a DataFrame with several columns than the pd.Series will have a value for the weighted mean of each of the columns.
        If inf_rm=False then:
            If v has Inf then the output will be Inf.
            If w has Inf then the output will be np.nan.
    """
    v, w = _prepare_weighted_stat_args(v, w, inf_rm)
    means = weighted_mean(v, w)
    v1 = w.sum()
    v2 = (w**2).sum()
    return (v1 / (v1**2 - v2)) * ((v - means) ** 2).multiply(w, axis="index").sum()


def weighted_sd(
    v: Union[
        List,
        pd.Series,
        pd.DataFrame,
        np.matrix,
    ],
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    inf_rm: bool = False,
) -> pd.Series:
    """Calculate the sample weighted standard deviation

    See :func:`weighted_var` for details.

    Args:
        v (Union[ List, pd.Series, pd.DataFrame, np.matrix, ]): Values.
        w (Union[ List, pd.Series, np.ndarray, None, ], optional): Weights. Defaults to None.
        inf_rm (bool, optional): Remove inf. Defaults to False.

    Returns:
        pd.Series: np.sqrt of :func:`weighted_var` (np.float64)
    """
    return np.sqrt(weighted_var(v, w, inf_rm))


def weighted_quantile(
    v: Union[
        List,
        pd.Series,
        pd.DataFrame,
        np.ndarray,
        np.matrix,
    ],
    quantiles: Union[
        List,
        pd.Series,
        np.ndarray,
    ],
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    inf_rm: bool = False,
) -> pd.DataFrame:
    """
    Calculates the weighted quantiles (q) of values (v) based on weights (w).

    See :func:`_prepare_weighted_stat_args` for the pre-processing done to v and w.

    Based on :func:`statsmodels.stats.weightstats.DescrStatsW`.

    Args:
        v (Union[ List, pd.Series, pd.DataFrame, np.array, np.matrix, ]): values to get the weighted quantiles for.
        quantiles (Union[ List, pd.Series, ]): the quantiles to calculate.
        w (Union[ List, pd.Series, np.array, ] optional): weights. Defaults to None.
        inf_rm (bool, optional): whether to remove infinite (from weights or values) from the computation. Defaults to False.

    Returns:
        pd.DataFrame: The index (names p) has the values from quantiles. The columns are based on v:
            If it's a pd.Series it's one column, if it's a pd.DataFrame with several columns, than each column
            in the output corrosponds to the column in v.
    """

    v, w = _prepare_weighted_stat_args(v, w, inf_rm)

    v = np.array(v)
    w = np.array(w)
    quantiles = np.array(quantiles)

    d = DescrStatsW(v.astype(float), weights=w)
    return d.quantile(quantiles)


def descriptive_stats(
    df: pd.DataFrame,
    weights: Union[
        List,
        pd.Series,
        np.ndarray,
        None,
    ] = None,
    stat: Literal["mean", "std", "var_of_mean", "ci_of_mean", "..."] = "mean",
    # relevant only if stat is None
    weighted: bool = True,
    # relevant only if we have non-numeric columns and we want to use model_matrix on them
    numeric_only: bool = False,
    add_na: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """Computes weighted statistics (e.g.: mean, std) on a DataFrame

    This function gets a DataFrame + weights and apply some weighted aggregation function (mean, std, or DescrStatsW).
    The main benefit of the function is that if the DataFrame includes non-numeric columns, then descriptive_stats will first
    run :func:`model_matrix` to create some numeric dummary variable that will then be processed.

    Args:
        df (pd.DataFrame): Some DataFrame to get stats (mean, std, etc.) for.
        weights (Union[ List, pd.Series, np.ndarray, ], optional): Weights to apply for the computation. Defaults to None.
        stat (Literal["mean", "std", "var_of_mean", ...], optional): Which statistic to calculate on the data.
            If mean - uses :func:`weighted_mean` (with inf_rm=True)
            If std - uses :func:`weighted_sd` (with inf_rm=True)
            If var_of_mean - uses :func:`var_of_weighted_mean` (with inf_rm=True)
            If ci_of_mean - uses :func:`ci_of_weighted_mean` (with inf_rm=True)
            If something else - tries to use :func:`statsmodels.stats.weightstats.DescrStatsW`.
                This supports stat such as: std_mean, sum_weights, nobs, etc. See function documentation to see more.
                (while removing mutual nan using :func:`rm_mutual_nas`)
            Defaults to "mean".
        weighted (bool, optional): If stat is not "mean" or "std", if to use the weights with the DescrStatsW function.
            Defaults to True.
        numeric_only (bool, optional): Should the statistics be computed only on numeric columns?
            If True - then non-numeric columns will be omitted.
            If False - then :func:`model_matrix` (with no formula argument) will be used to transfer non-numeric columns to dummy variables.
            Defaults to False.
        add_na (bool, optional): Passed to :func:`model_matrix`.
            Relevant only if numeric_only == False and df has non-numeric columns.
            Defaults to True.
        **kwargs: extra args to be passed to functions (e.g.: ci_of_weighted_mean)

    Returns:
        pd.DataFrame: Returns pd.DataFrame of the output (based on stat argument), for each of the columns in df.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots.weighted_stats import descriptive_stats, weighted_mean, weighted_sd

            # Without weights
            x = [1, 2, 3, 4]
            print(descriptive_stats(pd.DataFrame(x), stat="mean"))
            print(np.mean(x))
            print(weighted_mean(x))
                #     0
                # 0  2.5
                # 2.5
                # 0    2.5
                # dtype: float64

            print(descriptive_stats(pd.DataFrame(x), stat="var_of_mean"))
                #         0
                # 0  0.3125

            print(descriptive_stats(pd.DataFrame(x), stat="std"))
            print(weighted_sd(x))
            x2 = pd.Series(x)
            print(np.sqrt( np.sum((x2 - x2.mean())**2) / (len(x)-1) ))
                #         0
                # 0  1.290994
                # 0    1.290994
                # dtype: float64
                # 1.2909944487358056
                # Notice that it is different from
                # print(np.std(x))
                # which gives: 1.118033988749895
                # Which is the MLE (i.e.: biased, dividing by n and not n-1) estimator for std:
                # (np.sqrt( np.sum((x2 - x2.mean())**2) / (len(x)) ))

            x2 = pd.Series(x)
            tmp_sd = np.sqrt(np.sum((x2 - x2.mean()) ** 2) / (len(x) - 1))
            tmp_se = tmp_sd / np.sqrt(len(x))
            print(descriptive_stats(pd.DataFrame(x), stat="std_mean").iloc[0, 0])
            print(tmp_se)
                # 0.6454972243679029
                # 0.6454972243679028

            # Weighted results
            x, w = [1, 2, 3, 4], [1, 2, 3, 4]
            print(descriptive_stats(pd.DataFrame(x), w, stat="mean"))
                #      0
                # 0  3.0
            print(descriptive_stats(pd.DataFrame(x), w, stat="std"))
                #           0
                # 0  1.195229
            print(descriptive_stats(pd.DataFrame(x), w, stat="std_mean"))
                #           0
                # 0  0.333333
            print(descriptive_stats(pd.DataFrame(x), w, stat="var_of_mean"))
                #       0
                # 0  0.24
            print(descriptive_stats(pd.DataFrame(x), w, stat="ci_of_mean", conf_level = 0.99, round_ndigits=3))
                #                 0
                # 0  (1.738, 4.262)

    """
    if len(df.select_dtypes(np.number).columns) != len(df.columns):
        # If we have non-numeric columns, and want faster results,
        # then we can set numeric_only == True.
        # This will skip the model_matrix computation for non-numeric variables.
        if numeric_only:
            # TODO: (p2) does this check takes a long time?
            #       if it does - then maybe add an option of numeric_only = None
            #       to just use df as is.
            df = df.select_dtypes(include=[np.number])
        else:
            # TODO: add the ability to pass formula argument to model_matrix
            df = model_matrix(  # pyre-ignore[9]: this uses the DataFrame onlyÍ
                df, add_na=add_na, return_type="one"
            )["model_matrix"]

    if stat == "mean":
        return weighted_mean(df, weights, inf_rm=True).to_frame().transpose()
    elif stat == "std":
        return weighted_sd(df, weights, inf_rm=True).to_frame().transpose()
    elif stat == "var_of_mean":
        return var_of_weighted_mean(df, weights, inf_rm=True).to_frame().transpose()
    elif stat == "ci_of_mean":
        conf_level = kwargs.get("conf_level", 0.95)
        round_ndigits = kwargs.get("round_ndigits", None)
        return (
            ci_of_weighted_mean(
                df,
                weights,
                inf_rm=True,
                conf_level=conf_level,
                round_ndigits=round_ndigits,
            )
            .to_frame()
            .transpose()
        )

    # TODO: (p2) check which input DescrStatsW takes, not sure we need to run the next two lines.
    if weights is not None:
        weights = pd.Series(weights)
    # TODO: (p2) consider to remove the "weighted" argument, and just use
    #       whatever is passed to "weights" (if None, than make sure it's replaced with 1s)
    #  Fallback to statsmodels function
    _check_weights_are_valid(weights)
    wdf = {}
    for c in df.columns.values:
        df_c, w = rm_mutual_nas(df.loc[:, c], weights)
        wdf[c] = [getattr(DescrStatsW(df_c, w if weighted else None), stat)]
    return pd.DataFrame(wdf)


def relative_frequency_table(
    df: Union[pd.DataFrame, pd.Series],
    column: Optional[str] = None,
    w: Optional[pd.Series] = None,
) -> pd.DataFrame:
    """Creates a relative frequency table by aggregating over a categorical column (`column`) - optionally weighted by `w`.
    I.e.: produce the proportion (or weighted proportion) of rows that appear in each category, relative to the total number of rows (or sum of weights).
    See: https://en.wikipedia.org/wiki/Frequency_(statistics)#Types.

    Used in plotting functions.

    Args:
        df (pd.DataFrame): A DataFrame with categorical columns, or a pd.Series of the grouping column.
        column (Optional[str]): The name of the column to be aggregated.
            If None (default), then it takes the first column of df (if pd.DataFrame), or just uses as is (if pd.Series)
        w (Optional[pd.Series], optional): Optional weights to use when aggregating the relative proportions.
            If None than assumes weights is 1 for all rows. The defaults is None.

    Returns:
        pd.DataFrame: a pd.DataFrame with columns:
            - `column`, the aggregation variable, and,
            -  'prop', the aggregated (weighted) proportion of rows from each group in 'column'.

    Examples:
        ::

            from balance.stats_and_plots.weighted_stats import relative_frequency_table
            import pandas as pd

            df = pd.DataFrame({
                'group': ('a', 'b', 'c', 'c'),
                'v1': (1, 2, 3, 4),
            })
            print(relative_frequency_table(df, 'group'))
                #     group  prop
                #   0     a  0.25
                #   1     b  0.25
                #   2     c  0.50
            print(relative_frequency_table(df, 'group', pd.Series((2, 1, 1, 1),)))
                #     group  prop
                #   0     a   0.4
                #   1     b   0.2
                #   2     c   0.4

            # Using a pd.Series:
            a_series = df['group']
            print(relative_frequency_table(a_series))
                #   group  prop
                # 0     a  0.25
                # 1     b  0.25
                # 2     c  0.50
    """
    _check_weights_are_valid(w)

    if not (w is None or isinstance(w, pd.Series)):
        raise TypeError("argument `w` must be a pandas Series or None")
    if w is None:
        w = pd.Series(np.ones(df.shape[0]))

    # pyre-ignore[6]: this is a pyre bug. str inherits from hashable, and .rename works fine.
    w = w.rename("Freq")

    if column is None:
        if isinstance(df, pd.DataFrame):
            column = df.columns.values[0]
        elif isinstance(df, pd.Series):
            if df.name is None:
                df.name = "group"
            column = df.name
        else:
            raise TypeError("argument `df` must be a pandas DataFrame or Series")

    # pyre-ignore[6]: this is a bug. pd.concat can deal with a DataFrame and a Series.
    relative_frequency_table_data = pd.concat((df, w), axis=1)

    relative_frequency_table_data = relative_frequency_table_data.groupby(
        column, as_index=False
    ).sum()
    relative_frequency_table_data["prop"] = relative_frequency_table_data["Freq"] / sum(
        relative_frequency_table_data["Freq"]
    )

    return relative_frequency_table_data.loc[:, (column, "prop")]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import annotations

import collections
import inspect
import logging
from copy import deepcopy
from typing import Callable, Dict, List, Literal, Optional, Union

import numpy as np
import pandas as pd

from balance import adjustment as balance_adjustment, util as balance_util
from balance.stats_and_plots import weights_stats

from balance.stats_and_plots.weighted_comparisons_stats import outcome_variance_ratio
from balance.typing import FilePathOrBuffer

from IPython.lib.display import FileLink


logger: logging.Logger = logging.getLogger(__package__)


class Sample:
    """
    A class used to represent a sample.

    Sample is the main object of balance. It contains a dataframe of unit's observations,
    associated with id and weight.

    Attributes
    ----------
    id_column : pd.Series
        a column representing the ids of the units in sample
    weight_column : pd.Series
        a column representing the weights of the units in sample
    """

    # The following attributes are updated when initiating Sample using Sample.from_frame
    _df = None
    id_column = None
    _outcome_columns = None
    weight_column = None
    _links = None
    _adjustment_model = None
    _df_dtypes = None

    def __init__(self) -> None:
        # The following checks if the call to Sample() was initiated inside the class itself using from_frame, or outside of it
        # If the call was made internally, it will enable the creation of an instance of the class.
        # This is used when from_frame calls `sample = Sample()`. Keeping the full stack allows this also to work by a child of Sample.
        # If Sample() is called outside of the class structure, it will return the NotImplementedError error.
        try:
            calling_functions = [x.function for x in inspect.stack()]
        except Exception:
            raise NotImplementedError(
                "cannot construct Sample class directly... yet (only by invoking Sample.from_frame(...)"
            )

        if "from_frame" not in calling_functions:
            raise NotImplementedError(
                "cannot construct Sample class directly... yet (only by invoking Sample.from_frame(...)"
            )
        pass

    def __repr__(self: "Sample") -> str:
        return (
            f"({self.__class__.__module__}.{self.__class__.__qualname__})\n"
            f"{self.__str__()}"
        )

    def __str__(self: "Sample", pkg_source: str = __package__) -> str:
        is_adjusted = self.is_adjusted() * "Adjusted "
        n_rows = self._df.shape[0]
        n_variables = self._covar_columns().shape[1]
        has_target = self.has_target() * " with target set"
        adjustment_method = (
            " using " + self.model()["method"]  # pyre-ignore[16]
            # (None is eliminated by if statement)
            if self.model() is not None
            else ""
        )
        variables = ",".join(self._covar_columns_names())
        id_column_name = self.id_column.name if self.id_column is not None else "None"
        weight_column_name = (
            self.weight_column.name if self.weight_column is not None else "None"
        )
        outcome_column_names = (
            ",".join(self._outcome_columns.columns.tolist())
            if self._outcome_columns is not None
            else "None"
        )

        desc = f"""
        {is_adjusted}{pkg_source} Sample object{has_target}{adjustment_method}
        {n_rows} observations x {n_variables} variables: {variables}
        id_column: {id_column_name}, weight_column: {weight_column_name},
        outcome_columns: {outcome_column_names}
        """

        if self.has_target():
            common_variables = balance_util.choose_variables(
                self, self._links["target"], variables=None
            )
            target_str = self._links["target"].__str__().replace("\n", "\n\t")
            n_common = len(common_variables)
            common_variables = ",".join(common_variables)
            desc += f"""
            target:
                 {target_str}
            {n_common} common variables: {common_variables}
            """
        return desc

    ################################################################################
    #  Public API
    ################################################################################

    @classmethod
    def from_frame(
        cls: type["Sample"],
        df: pd.DataFrame,
        id_column: Optional[str] = None,
        outcome_columns: Optional[Union[list, tuple, str]] = None,
        weight_column: Optional[str] = None,
        check_id_uniqueness: bool = True,
        standardize_types: bool = True,
        use_deepcopy: bool = True,
    ) -> "Sample":
        """
        Create a new Sample object.

        NOTE that all integer columns will be converted by defaults into floats. This behavior can be turned off
        by setting standardize_types argument to False.
        The reason this is done by default is because of missing value handling combined with balance current lack of support
        for pandas Integer types:
            1. Native numpy integers do not support missing values (NA), while pandas Integers do,
            as well numpy floats. Also,
            2. various functions in balance do not support pandas Integers, while they do support numpy floats.
            3. Hence, since some columns might have missing values, the safest solution is to just convert all integers into numpy floats.

        The id_column is stored as a string, even if the input is an integer.

        Args:
            df (pd.DataFrame): containing the sample's data
            id_column (Optional, Optional[str]): the column of the df which contains the respondent's id
            (should be unique). Defaults to None.
            outcome_columns (Optional, Optional[Union[list, tuple, str]]): names of columns to treat as outcome
            weight_column (Optional, Optional[str]): name of column to treat as weight. If not specified, will
                be guessed (either "weight" or "weights"). If not found, a new column will be created ("weight") and filled with 1.0.
            check_id_uniqueness (Optional, bool): Whether to check if ids are unique. Defaults to True.
            standardize_types (Optional, bool): Whether to standardize types. Defaults to True.
                Int64/int64 -> float64
                Int32/int32 -> float64
                string -> object
                pandas.NA -> numpy.nan (within each cell)
                This is slightly memory intensive (since it copies the data twice),
                but helps keep various functions working for both Int64 and Int32 input columns.
            use_deepcopy (Optional, bool): Whether to have a new df copy inside the sample object.
                If False, then when the sample methods update the internal df then the original df will also be updated.
                Defaults to True.

        Returns:
            Sample: a sample object
        """
        # Inititate a Sample() class, inside a from_frame constructor
        sample = cls()

        sample._df_dtypes = df.dtypes

        if use_deepcopy:
            sample._df = deepcopy(df)
        else:
            sample._df = df

        # id column
        id_column = balance_util.guess_id_column(df, id_column)
        if any(sample._df[id_column].isnull()):
            raise ValueError("Null values are not allowed in the id_column")
        if not set(map(type, sample._df[id_column].tolist())) == {  # pyre-fixme[6] ???
            str
        }:
            logger.warning("Casting id column to string")
            sample._df.loc[:, id_column] = sample._df.loc[:, id_column].astype(str)

        if (check_id_uniqueness) and (
            sample._df[id_column].nunique() != len(sample._df[id_column])
        ):
            raise ValueError("Values in the id_column must be unique")
        sample.id_column = sample._df[id_column]

        # TODO: in the future, if we could have all functions work with the original data types, that would be better.
        if standardize_types:
            # Move from some pandas Integer types to numpy float types.
            # NOTE: The rationale is that while pandas integers support missing values,
            #       numpy float types do (storing it as np.nan).
            #       Furthermore, other functions in the package don't handle pandas Integer objects well, so
            #       they must be converted to numpy integers (if they have no missing values).
            #       But since we can't be sure that none of the various objects with the same column will not have NAs,
            #       we just convert them all to np.float (either 32 or 64).
            #       For more details, see: https://stackoverflow.com/a/53853351
            # This line is after the id_column is set, so to make sure that the conversion happens after it is stored as a string.
            # Move from Int64Dtype() to dtype('int64'):

            # TODO: convert all numeric values (no matter what the original is) to "float64"?
            #       (Instead of mentioning all different types)
            #       using is_numeric_dtype: https://pandas.pydata.org/docs/reference/api/pandas.api.types.is_numeric_dtype.html?highlight=is_numeric_dtype#pandas.api.types.is_numeric_dtype
            #       Also, consider using
            #       https://pandas.pydata.org/docs/reference/api/pandas.api.types.is_string_dtype.html
            #       or https://pandas.pydata.org/docs/reference/api/pandas.api.types.is_object_dtype.html
            #       from non-numeric.
            #       e.d.: balance/util.py?lines=512.
            #           for x in df.columns:
            #               if (is_numeric_dtype(df[x])) and (not is_bool_dtype(df[x])):
            #                   df[x] = df[x].astype("float64")
            input_type = ["Int64", "Int32", "int64", "int32", "int16", "int8", "string"]
            output_type = [
                "float64",
                "float32",  # This changes Int32Dtype() into dtype('int32') (from pandas to numpy)
                "float64",
                "float32",
                "float16",
                "float16",  # Using float16 since float8 doesn't exist, see: https://stackoverflow.com/a/40507235/256662
                "object",
            ]
            for i_input, i_output in zip(input_type, output_type):
                sample._df = balance_util._pd_convert_all_types(
                    sample._df, i_input, i_output
                )

            # Replace any pandas.NA with numpy.nan:
            sample._df = sample._df.fillna(np.nan)

            balance_util._warn_of_df_dtypes_change(
                sample._df_dtypes,
                sample._df.dtypes,
                "df",
                "sample._df",
            )

        # weight column
        if weight_column is None:
            if "weight" in sample._df.columns:
                logger.warning("Guessing weight column is 'weight'")
                weight_column = "weight"
            elif "weights" in sample._df.columns:
                logger.warning("Guessing weight column is 'weights'")
                weight_column = "weights"
            else:
                # TODO: The current default when weights are not available is "weight", while the method in balanceDF is called "weights",
                #       and the subclass is called BalanceWeightsDF (with and 's' at the end)
                #       In the future, it would be better to be more consistent and use the same name for all variations (e.g.: "weight").
                #       Unless, we move to use more weights columns, and this method could be used to get all of them.
                logger.warning(
                    "No weights passed. Adding a 'weight' column and setting all values to 1"
                )
                weight_column = "weight"
                sample._df.loc[:, weight_column] = 1

        sample.weight_column = sample._df[weight_column]

        # outcome columns
        if outcome_columns is None:
            sample._outcome_columns = None
        else:
            if isinstance(outcome_columns, str):
                outcome_columns = [outcome_columns]
            try:
                sample._outcome_columns = sample._df.loc[:, outcome_columns]
            except KeyError:
                _all_columns = sample._df.columns.values.tolist()
                raise ValueError(
                    f"outcome columns {outcome_columns} not in df columns {_all_columns}"
                )

        sample._links = collections.defaultdict(list)
        return sample

    ####################
    # Class base methods
    ####################
    @property
    def df(self: "Sample") -> pd.DataFrame:
        """Produce a DataFrame (of the self) from a Sample object.

        Args:
            self (Sample): Sample object.

        Returns:
            pd.DataFrame: with id_columns, and the df values of covars(), outcome() and weights() of the self in the Sample object.
        """
        return pd.concat(
            (
                self.id_column,
                self.covars().df if self.covars() is not None else None,
                self.outcomes().df if self.outcomes() is not None else None,
                self.weights().df if self.weights() is not None else None,
            ),
            axis=1,
        )

    def outcomes(
        self: "Sample",
    ):  # -> "Optional[Type[BalanceOutcomesDF]]" (not imported due to circular dependency)
        """
        Produce a BalanceOutcomeDF from a Sample object.
        See :class:BalanceOutcomesDF.

        Args:
            self (Sample): Sample object.

        Returns:
            BalanceOutcomesDF or None
        """
        if self._outcome_columns is not None:
            # NOTE: must import here so to avoid circular dependency
            from balance.balancedf_class import BalanceOutcomesDF

            return BalanceOutcomesDF(self)
        else:
            return None

    def weights(
        self: "Sample",
    ):  # -> "Optional[Type[BalanceWeightsDF]]" (not imported due to circular dependency)
        """
        Produce a BalanceWeightsDF from a Sample object.
        See :class:BalanceWeightsDF.

        Args:
            self (Sample): Sample object.

        Returns:
            BalanceWeightsDF
        """
        # NOTE: must import here so to avoid circular dependency
        from balance.balancedf_class import BalanceWeightsDF

        return BalanceWeightsDF(self)

    def covars(
        self: "Sample",
    ):  # -> "Optional[Type[BalanceCovarsDF]]" (not imported due to circular dependency)
        """
        Produce a BalanceCovarsDF from a Sample object.
        See :class:BalanceCovarsDF.

        Args:
            self (Sample): Sample object.

        Returns:
            BalanceCovarsDF
        """
        # NOTE: must import here so to avoid circular dependency
        from balance.balancedf_class import BalanceCovarsDF

        return BalanceCovarsDF(self)

    def model(
        self: "Sample",
    ) -> Optional[Dict]:
        """
        Returns the name of the model used to adjust Sample if adjusted.
        Otherwise returns None.

        Args:
            self (Sample): Sample object.

        Returns:
            str or None: name of model used for adjusting Sample
        """
        if hasattr(self, "_adjustment_model"):
            return self._adjustment_model
        else:
            return None

    def model_matrix(self: "Sample") -> pd.DataFrame:
        """
        Returns the model matrix of sample using :func:`model_matrix`,
        while adding na indicator for null values (see :func:`add_na_indicator`).

        Returns:
            pd.DataFrame: model matrix of sample
        """
        res = balance_util.model_matrix(self, add_na=True)["sample"]
        return res  # pyre-ignore[7]: ["sample"] only chooses the DataFrame

    ############################################
    # Adjusting and adapting weights of a sample
    ############################################
    def adjust(
        self: "Sample",
        target: Optional["Sample"] = None,
        method: Union[
            Literal["cbps", "ipw", "null", "poststratify", "rake"], Callable
        ] = "ipw",
        *args,
        **kwargs,
    ) -> "Sample":
        """
        Perform adjustment of one sample to match another.
        This function returns a new sample.

        Args:
            target (Optional["Sample"]): Second sample object which should be matched.
                If None, the set target of the object is used for matching.
            method (str): method for adjustment: cbps, ipw, null, poststratify, rake

        Returns:
            Sample: an adjusted Sample object
        """
        if target is None:
            self._no_target_error()
            target = self._links["target"]

        new_sample = deepcopy(self)
        if type(method) == str:
            adjustment_function = balance_adjustment._find_adjustment_method(method)
        elif callable(method):
            adjustment_function = method
        else:
            raise ValueError("Method should be one of existing weighting methods")

        adjusted = adjustment_function(
            sample_df=self.covars().df,
            sample_weights=self.weight_column,
            target_df=target.covars().df,
            target_weights=target.weight_column,
            *args,
            **kwargs,
        )
        new_sample.set_weights(adjusted["weight"])
        new_sample._adjustment_model = adjusted["model"]
        new_sample._links["unadjusted"] = self
        new_sample._links["target"] = target

        return new_sample

    def set_weights(self, weights: Optional[Union[pd.Series, float]]) -> None:
        """
        Adjusting the weights of a Sample object.
        This will overwrite the weight_column of the Sample.
        Note that the weights are assigned by index if weights is a pd.Series
        (of Sample.df and weights series)

        Args:
            weights (Optional[Union[pd.Series, float]]): Series of weights to add to sample.
                If None or float values, the same weight (or None) will be assigned to all units.

        Returns:
            None, but adapting the Sample weight column to weights
        """
        if isinstance(weights, pd.Series):
            if not all(idx in weights.index for idx in self.df.index):
                logger.warning(
                    """Note that not all Sample units will be assigned weights,
                    since weights are missing some of the indices in Sample.df"""
                )
        self._df.loc[:, self.weight_column.name] = weights
        self.weight_column = self._df[self.weight_column.name]

    ####################################
    # Handling links to other dataframes
    ####################################
    def set_unadjusted(self, second_sample: "Sample") -> "Sample":
        """
        Used to set the unadjusted link to Sample.
        This is useful in case one wants to compare two samples.

        Args:
            second_sample (Sample): A second Sample to be set as unadjusted of Sample.

        Returns:
            Sample: a new copy of Sample with unadjusted link attached to the self object.
        """
        if isinstance(second_sample, Sample):
            newsample = deepcopy(self)
            newsample._links["unadjusted"] = second_sample
            return newsample
        else:
            raise TypeError(
                "set_unadjusted must be called with second_sample argument of type Sample"
            )

    def is_adjusted(self) -> bool:
        """Check if a Sample object is adjusted and has target attached

        Returns:
            bool: whether the Sample is adjusted or not.
        """
        return ("unadjusted" in self._links) and ("target" in self._links)

    def set_target(self, target: "Sample") -> "Sample":
        """
        Used to set the target linked to Sample.

        Args:
            target (Sample): A Sample object to be linked as target

        Returns:
            Sample: new copy of Sample with target link attached
        """
        if isinstance(target, Sample):
            newsample = deepcopy(self)
            newsample._links["target"] = target
            return newsample
        else:
            raise ValueError("A target, a Sample object, must be specified")

    def has_target(self) -> bool:
        """
        Check if a Sample object has target attached.

        Returns:
            bool: whether the Sample has target attached
        """
        return "target" in self._links

    ##############################
    # Metrics for adjusted samples
    ##############################
    def covar_means(self: "Sample") -> pd.DataFrame:
        """
        Compare the means of covariates (after using :func:`BalanceDF.model_matrix`) before and after adjustment as compared with target.

        Args:
            self (Sample): A Sample object produces after running :func:`Sample.adjust`.
                It should include 3 components: "unadjusted", "adjusted", "target".

        Returns:
            pd.DataFrame: A DataFrame with 3 columns ("unadjusted", "adjusted", "target"),
            and a row for each feature of the covariates.
            The cells show the mean value. For categorical features, they are first transformed into the one-hot encoding.
            For these columns, since they are all either 0 or 1, their means should be interpreted as proportions.

        Examples:
            ::

                from balance import Sample
                import pandas as pd

                s = Sample.from_frame(
                    pd.DataFrame(
                        {"a": (0, 1, 2), "c": ("a", "b", "c"), "o": (1,3,5), "id": (1, 2, 3)}
                    ),
                    outcome_columns=("o"),
                )
                s_adjusted = s.set_target(s).adjust(method = 'null')
                print(s_adjusted.covar_means())

                    # source  unadjusted  adjusted    target
                    # a         1.000000  1.000000  1.000000
                    # c[a]      0.333333  0.333333  0.333333
                    # c[b]      0.333333  0.333333  0.333333
                    # c[c]      0.333333  0.333333  0.333333
        """
        self._check_if_adjusted()

        means = self.covars().mean()
        means = (
            means.rename(index={"self": "adjusted"})
            .reindex(["unadjusted", "adjusted", "target"])
            .transpose()
        )

        return means

    def design_effect(self) -> np.float64:
        """
        Return the design effect of the weights of Sample. Uses :func:`weights_stats.design_effect`.

        Args:
            self (Sample): A Sample object

        Returns:
            np.float64: Design effect
        """
        return weights_stats.design_effect(self.weight_column)

    def design_effect_prop(self) -> np.float64:
        """
        Return the relative difference in design effect of the weights of the unadjusted sample and the adjusted sample.
        I.e. (Deff of adjusted - Deff of unadjusted) / Deff of unadjusted.
        Uses :func:`weights_stats.design_effect`.

        Args:
            self (Sample): A Sample object produces after running :func:`Sample.adjust`.
                It should include 3 components: "unadjusted", "adjusted", "target".

        Returns:
            np.float64: relative difference in design effect.
        """
        self._check_if_adjusted()
        deff_unadjusted = self._links["unadjusted"].design_effect()
        deff_adjusted = self.design_effect()

        return (deff_adjusted - deff_unadjusted) / deff_unadjusted

    # TODO: add unittest for this function
    def plot_weight_density(self) -> None:
        """Plot the density of weights of Sample.

        Examples:
            ::

                import numpy as np
                import pandas as pd
                from balance.sample_class import Sample


                np.random.seed(123)
                df = pd.DataFrame(
                    {
                        "a": np.random.uniform(size=100),
                        "c": np.random.choice(
                            ["a", "b", "c", "d"],
                            size=100,
                            replace=True,
                            p=[0.01, 0.04, 0.5, 0.45],
                        ),
                        "id": range(100),
                        "weight": np.random.uniform(size=100) + 0.5,
                    }
                )

                a = Sample.from_frame(df)
                sample.weights().plot()
                # The same as:
                sample.plot_weight_density()
        """
        self.weights().plot()

    ##########################################
    # Metrics for outcomes of adjusted samples
    ##########################################
    def outcome_sd_prop(self) -> pd.Series:
        """
        Return the difference in outcome weighted standard deviation (sd) of the unadjusted
        sample and the adjusted sample, relative to the unadjusted weighted sd.
        I.e. (weighted sd of adjusted - weighted sd of unadjusted) / weighted sd  of unadjusted.
        Uses :func:`BalanceDF.weighted_stats.weighted_sd`.

        Args:
            self (Sample): A Sample object produces after running :func:`Sample.adjust`.
                It should include 3 components: "unadjusted", "adjusted", "target".

        Returns:
            pd.Series: (np.float64) relative difference in outcome weighted standard deviation.
        """
        self._check_if_adjusted()
        self._check_outcomes_exists()

        outcome_std = self.outcomes().std()
        adjusted_outcome_sd = outcome_std.loc["self"]
        unadjusted_outcome_sd = outcome_std.loc["unadjusted"]

        return (adjusted_outcome_sd - unadjusted_outcome_sd) / unadjusted_outcome_sd

    def outcome_variance_ratio(self: "Sample") -> pd.Series:
        """The empirical ratio of variance of the outcomes before and after weighting.

        See :func:`outcome_variance_ratio` for details.

        Args:
            self (Sample): A Sample object produces after running :func:`Sample.adjust`.
                It should include 3 components: "unadjusted", "adjusted", "target".

        Returns:
             pd.Series: (np.float64) A series of calculated ratio of variances for each outcome.
        """
        return outcome_variance_ratio(
            self.outcomes().df,
            self._links["unadjusted"].outcomes().df,
            self.weights().df["weight"],
            self._links["unadjusted"].weights().df["weight"],
        )

    # TODO: Add a method that plots the distribution of the outcome (adjusted v.s. unadjusted
    #       if adjusted, and only unadjusted otherwise)

    ##############################################
    # Summary of metrics and diagnostics of Sample
    ##############################################
    def summary(self) -> str:
        """
        Provides a summary of covariate balance, design effect and model properties (if applicable)
        of a sample.

        For more details see: :func:`BalanceDF.asmd`, :func:`BalanceDF.asmd_improvement`
        and :func:`weights_stats.design_effect`

        Returns:
            str: a summary description of properties of an adjusted sample.
        """
        # asmd
        if self.is_adjusted() or self.has_target():
            asmd = self.covars().asmd()
            n_asmd_covars = len(
                asmd.columns.values[asmd.columns.values != "mean(asmd)"]
            )

        # asmd improvement
        if self.is_adjusted():
            asmd_before = asmd.loc["unadjusted", "mean(asmd)"]
            asmd_improvement = 100 * self.covars().asmd_improvement()

        if self.has_target():
            asmd_now = asmd.loc["self", "mean(asmd)"]

        # design effect
        design_effect = self.design_effect()

        # model performance
        if self.model() is not None:
            if (
                self.model()["method"]  # pyre-ignore[16]
                # (None is eliminated by if statement)
                == "ipw"
            ):
                model_summary = (
                    "Model proportion deviance explained: {dev_exp:.3f}".format(
                        dev_exp=self.model()["perf"]["prop_dev_explained"][0]
                    )
                )
            else:
                # TODO: add model performance for other types of models
                model_summary = None
        else:
            model_summary = None

        out = (
            (
                # pyre-fixme[61]: `asmd_improvement` is undefined, or not always
                #  defined.
                f"Covar ASMD reduction: {asmd_improvement:.1f}%, design effect: {design_effect:.3f}\n"
                if self.is_adjusted()
                else ""
            )
            # pyre-fixme[61]: `n_asmd_covars` is undefined, or not always defined.
            + (f"Covar ASMD ({n_asmd_covars} variables): " if self.has_target() else "")
            # pyre-fixme[61]: `asmd_before` is undefined, or not always defined.
            + (f"{asmd_before:.3f} -> " if self.is_adjusted() else "")
            # pyre-fixme[61]: `asmd_now` is undefined, or not always defined.
            + (f"{asmd_now:.3f}\n" if self.has_target() else "")
            + (
                f"Model performance: {model_summary}"
                if (model_summary is not None)
                else ""
            )
        )
        return out

    def diagnostics(self: "Sample") -> pd.DataFrame:
        # TODO: mention the other diagnostics
        # TODO: update/improve the wiki pages doc is linking to.
        # TODO: move explanation on weights normalization to some external page
        """
        Output a table of diagnostics about adjusted Sample object.

        size
        ======================
        All values in the "size" metrics are AFTER any rows/columns were filtered.
        So, for example, if we use respondents from previous days but filter them for diagnostics purposes, then
        sample_obs and target_obs will NOT include them in the counting. The same is true for sample_covars and target_covars.
        In the "size" metrics we have the following 'var's:
        - sample_obs - number of respondents
        - sample_covars -  number of covariates (main covars, before any transformations were used)
        - target_obs - number of users used to represent the target pop
        - target_covars - like sample_covars, but for target.

        weights_diagnostics
        ======================
        In the "weights_diagnostics" metric we have the following 'var's:
        - design effect (de), effective sample size (n/de), effective sample ratio (1/de). See also:
            - https://en.wikipedia.org/wiki/Design_effect
            - https://en.wikipedia.org/wiki/Effective_sample_size
        - sum
        - describe of the (normalized to sample size) weights (mean, median, std, etc.)
        - prop of the (normalized to sample size) weights that are below or above some numbers (1/2, 1, 2, etc.)
        - nonparametric_skew and weighted_median_breakdown_point

        Why is the diagnostics focused on weights normalized to sample size
        -------------------------------------------------------------------
        There are 3 well known normalizations of weights:
        1. to sum to 1
        2. to sum to target population
        3. to sum to n (sample size)

        Each one has their own merits:
        1. is good if wanting to easily calculate avg of some response var (then we just use sum(w*y) and no need for /sum(w))
        2. is good for sum of stuff. For example, how many people in the US use android? For this we'd like the weight of
            each person to represent their share of the population and then we just sum the weights of the people who use android in the survey.
        3. is good for understanding relative "importance" of a respondent as compared to the weights of others in the survey.
            So if someone has a weight that is >1 it means that this respondent (conditional on their covariates) was 'rare' in the survey,
            so the model we used decided to give them a larger weight to account for all the people like him/her that didn't answer.

        For diagnostics purposes, option 3 is most useful for discussing the distribution of the weights
        (e.g.: how many respondents got a weight >2 or smaller <0.5).
        This is a method (standardized  across surveys) to helping us identify how many of the respondents are "dominating"
        and have a large influence on the conclusion we draw from the survey.

        model_glance
        ======================
        Properties of the model fitted, depends on the model used for weighting.

        covariates ASMD
        ======================
        Includes covariates ASMD before and after adjustment (per level of covariate and aggregated) and the ASMD improvement.

        Args:
            self (Sample): only after running an adjustment with Sample.adjust.

        Returns:
            pd.DataFrame: with 3 columns: ("metric", "val", "var"),
                indicating various tracking metrics on the model.
        """
        logger.info("Starting computation of diagnostics of the fitting")
        self._check_if_adjusted()
        diagnostics = pd.DataFrame(columns=("metric", "val", "var"))

        # ----------------------------------------------------
        # Properties of the Sample object (dimensions of the data)
        # ----------------------------------------------------
        n_sample_obs, n_sample_covars = self.covars().df.shape
        n_target_obs, n_target_covars = self._links["target"].covars().df.shape

        diagnostics = pd.concat(
            (
                diagnostics,
                pd.DataFrame(
                    {
                        "metric": "size",
                        "val": [
                            n_sample_obs,
                            n_sample_covars,
                            n_target_obs,
                            n_target_covars,
                        ],
                        "var": [
                            "sample_obs",
                            "sample_covars",
                            "target_obs",
                            "target_covars",
                        ],
                    }
                ),
            )
        )

        # ----------------------------------------------------
        # Diagnostics on the weights
        # ----------------------------------------------------
        the_weights = self.weights().df.iloc[
            :, 0
        ]  # should be ['weight'], but this is more robust in case a user uses other names
        weights_diag_var = []
        weights_diag_value = []

        # adding design_effect and variations
        the_weights_de = weights_stats.design_effect(the_weights)
        weights_diag_var.extend(
            ["design_effect", "effective_sample_ratio", "effective_sample_size"]
        )
        weights_diag_value.extend(
            [the_weights_de, 1 / the_weights_de, len(the_weights) / the_weights_de]
        )

        # adding sum of weights, and then normalizing them to n (sample size)
        weights_diag_var.append("sum")
        weights_diag_value.append(the_weights.sum())

        the_weights = the_weights / the_weights.mean()  # normalize weights to sum to n.

        # adding basic summary statistics from describe:
        tmp_describe = the_weights.describe()
        weights_diag_var.extend(["describe_" + i for i in tmp_describe.index])
        weights_diag_value.extend(tmp_describe.to_list())
        # TODO: decide if we want more quantiles of the weights.

        # adding prop_above_and_below
        tmp_props = weights_stats.prop_above_and_below(the_weights)
        weights_diag_var.extend(
            tmp_props.index.to_list()  # pyre-ignore[16]: existing defaults make sure this output is pd.Series with relevant methods.
        )
        weights_diag_value.extend(
            tmp_props.to_list()  # pyre-ignore[16]: existing defaults make sure this output is pd.Series with relevant methods.
        )
        # TODO: decide if we want more numbers (e.g.: 2/3 and 3/2)

        # adding nonparametric_skew and weighted_median_breakdown_point
        weights_diag_var.append("nonparametric_skew")
        weights_diag_value.append(weights_stats.nonparametric_skew(the_weights))

        weights_diag_var.append("weighted_median_breakdown_point")
        weights_diag_value.append(
            weights_stats.weighted_median_breakdown_point(the_weights)
        )

        # Add all the weights_diagnostics to diagnostics
        diagnostics = pd.concat(
            (
                diagnostics,
                pd.DataFrame(
                    {
                        "metric": "weights_diagnostics",
                        "val": weights_diag_value,
                        "var": weights_diag_var,
                    }
                ),
            )
        )

        # ----------------------------------------------------
        # Diagnostics on the model
        # ----------------------------------------------------
        model = self.model()
        diagnostics = pd.concat(
            (
                diagnostics,
                pd.DataFrame(
                    {
                        "metric": "adjustment_method",
                        "val": (0,),
                        "var": model["method"],  # pyre-ignore[16]
                        # (None is eliminated by if statement)
                    }
                ),
            )
        )
        if model["method"] == "ipw":
            #  Scalar values from 'perf' key of dictionary
            fit_single_values = pd.concat(
                [
                    pd.DataFrame({"metric": "model_glance", "val": v, "var": k})
                    for k, v in model["fit"].items()
                    if (isinstance(v, np.ndarray) and v.shape == (1,))
                ]
            )
            diagnostics = pd.concat((diagnostics, fit_single_values))

            #  Extract glmnet output about this regularisation parameter
            lambda_ = model["lambda"]
            lambda_index = model["fit"]["lambdau"] == lambda_
            fit_values = pd.concat(
                [
                    pd.DataFrame(
                        {"metric": "model_glance", "val": v[lambda_index], "var": k}
                    )
                    for k, v in self.model()["fit"].items()
                    if (isinstance(v, np.ndarray) and v.shape) == lambda_index.shape
                ]
            )
            diagnostics = pd.concat((diagnostics, fit_values))

            #  Scalar values from 'perf' key of dictionary
            perf_single_values = pd.concat(
                [
                    pd.DataFrame({"metric": "model_glance", "val": v, "var": k})
                    for k, v in model["perf"].items()
                    if (isinstance(v, np.ndarray) and v.shape == (1,))
                ]
            )
            diagnostics = pd.concat((diagnostics, perf_single_values))

            # Model coefficients
            coefs = (
                model["perf"]["coefs"]
                .reset_index()
                .rename({0: "val", "index": "var"}, axis=1)
                .assign(metric="model_coef")
            )
            diagnostics = pd.concat((diagnostics, coefs))

        elif model["method"] == "cbps":
            beta_opt = pd.DataFrame(
                {"val": model["beta_optimal"], "var": model["X_matrix_columns"]}
            ).assign(metric="beta_optimal")
            diagnostics = pd.concat((diagnostics, beta_opt))

            metric = [
                "rescale_initial_result",
                "balance_optimize_result",
                "gmm_optimize_result_glm_init",
                "gmm_optimize_result_bal_init",
            ]
            metric = [x for x in metric for _ in range(2)]
            var = ["success", "message"] * 4
            val = [model[x][y] for (x, y) in zip(metric, var)]

            optimizations = pd.DataFrame({"metric": metric, "var": var, "val": val})
            diagnostics = pd.concat((diagnostics, optimizations))

        # TODO: add model diagnostics for other models

        # ----------------------------------------------------
        # Diagnostics on the covariates correction
        # ----------------------------------------------------
        asmds = self.covars().asmd()

        #  Per-covariate ASMDs
        covar_asmds = (
            asmds.transpose()
            .rename(
                {
                    "self": "covar_asmd_adjusted",
                    "unadjusted": "covar_asmd_unadjusted",
                    "unadjusted - self": "covar_asmd_improvement",
                },
                axis=1,
            )
            .reset_index()
            .melt(id_vars="index")
            .rename({"source": "metric", "value": "val", "index": "var"}, axis=1)
        )
        diagnostics = pd.concat((diagnostics, covar_asmds))

        #  Per-main-covariate ASMDs
        asmds_main = self.covars().asmd(aggregate_by_main_covar=True)
        covar_asmds_main = (
            asmds_main.transpose()
            .rename(
                {
                    "self": "covar_main_asmd_adjusted",
                    "unadjusted": "covar_main_asmd_unadjusted",
                    "unadjusted - self": "covar_main_asmd_improvement",
                },
                axis=1,
            )
            .reset_index()
            # TODO:
            # column index name is different here.
            # think again if that's the best default or not for
            # asmd(aggregate_by_main_covar = True)
            .rename({"main_covar_names": "index"}, axis=1)
            .melt(id_vars="index")
            .rename({"source": "metric", "value": "val", "index": "var"}, axis=1)
        )
        # sort covar_asmds_main to have mean(asmd) at the end of it (for when doing quick checks)
        covar_asmds_main = (
            covar_asmds_main.assign(
                has_mean_asmd=(covar_asmds_main["var"] == "mean(asmd)")
            )
            .sort_values(by=["has_mean_asmd", "var"])
            .drop(columns="has_mean_asmd")
        )
        diagnostics = pd.concat((diagnostics, covar_asmds_main))

        # ----------------------------------------------------
        # Diagnostics if there was an adjustment_failure
        # ----------------------------------------------------
        # This field is used in the cli and filled with an alternative value if needed.
        diagnostics = pd.concat(
            (
                diagnostics,
                pd.DataFrame({"metric": ("adjustment_failure",), "val": (0,)}),
            )
        )

        diagnostics = diagnostics.reset_index(drop=True)

        logger.info("Done computing diagnostics")
        return diagnostics

    ############################################
    # Column and rows modifiers - use carefully!
    ############################################
    def keep_only_some_rows_columns(
        self: "Sample",
        rows_to_keep: Optional[str] = None,
        columns_to_keep: Optional[List[str]] = None,
    ) -> "Sample":
        # TODO: split this into two functions (one for rows and one for columns)
        """
        This function returns a **copy** of the sample object
        after removing ALL columns from _df and _links objects
        (which includes unadjusted and target objects).

        This function is useful when wanting to calculate metrics, such as ASMD, but only on some of the features,
        or part of the observations.

        Args:
            self (Sample): a sample object (preferably after adjustment)
            rows_to_keep (Optional[str], optional): A string with a condition to eval (on some of the columns).
                This will run df.eval(rows_to_keep) which will return a pd.Series of bool by which
                we will filter the Sample object.
                This effects both the df of covars AND the weights column (weight_column)
                AND the outcome column (_outcome_columns), AND the id_column column.
                Input should be a boolean feature, or a condition such as: 'gender == "Female" & age >= 18'.
                Defaults to None.
            columns_to_keep (Optional[List[str]], optional): the covariates of interest.
                Defaults to None, which returns all columns.

        Returns:
            Sample: A copy of the original object. If both rows and columns to keep are None,
                returns the copied object unchanged.
                If some are not None, will update - first the rows - then the columns.
                This performs the transformation on both the sample's df and its linked dfs (unadjusted, target).
        """
        if (rows_to_keep is None) and (columns_to_keep is None):
            return self

        # Let's make sure to not ruin our old object:
        self = deepcopy(self)

        if rows_to_keep is not None:
            # let's filter the weights Series and then the df rows
            ss = self.df.eval(rows_to_keep)  # rows to keep after the subset # noqa
            logger.info(
                f"From self -> (rows_filtered/total_rows) = ({ss.sum()}/{len(ss)})"
            )
            # filter ids
            self.id_column = self.id_column[ss]
            # filter weights
            self.weight_column = self.weight_column[ss]
            # filter _df
            self._df = self._df[ss]
            # filter outcomes
            if self._outcome_columns is not None:
                self._outcome_columns = self._outcome_columns[ss]
            # filter links
            for k, v in self._links.items():
                try:
                    ss = v.df.eval(rows_to_keep)  # rows to keep after the subset # noqa
                    logger.info(
                        f"From {k} -> (rows_filtered/total_rows) = ({ss.sum()}/{len(ss)})"
                    )
                    v.id_column = v.id_column[ss]
                    v.weight_column = v.weight_column[ss]
                    v._df = v._df[ss]
                    if v._outcome_columns is not None:
                        v._outcome_columns = v._outcome_columns[ss]
                except pd.core.computation.ops.UndefinedVariableError:
                    # This can happen, for example, if the row filtering condition depends somehow on a feature that is
                    # in the sample but not in the _links. For example, if filtering over one of the
                    # outcome variables, it would filter out these rows from sample, but it wouldn't have this column to
                    # use in target. So this is meant to capture that when this happens the function won't fail but simply
                    # report it to the user.
                    logger.warning(
                        f"couldn't filter _links['{k}'] using {rows_to_keep}"
                    )

        if columns_to_keep is not None:
            if not (set(columns_to_keep) <= set(self.df.columns)):
                logger.warning(
                    "Note that not all columns_to_keep are in Sample. Only those exists are removed"
                )
            # let's remove columns...
            self._df = self._df.loc[:, self._df.columns.isin(columns_to_keep)]
            for v in self._links.values():
                v._df = v._df.loc[:, v._df.columns.isin(columns_to_keep)]

        return self

    ################
    # Saving results
    ################
    def to_download(self, tempdir: Optional[str] = None) -> FileLink:
        """Creates a downloadable link of the DataFrame of the Sample object.

        File name starts with tmp_balance_out_, and some random file name (using :func:`uuid.uuid4`).

        Args:
            self (Sample): Object.
            tempdir (Optional[str], optional): Defaults to None (which then uses a temporary folder using :func:`tempfile.gettempdir`).

        Returns:
            FileLink: Embedding a local file link in an IPython session, based on path. Using :func:FileLink.
        """
        return balance_util._to_download(self.df, tempdir)

    def to_csv(
        self, path_or_buf: Optional[FilePathOrBuffer] = None, **kwargs
    ) -> Optional[str]:
        """Write df with ids from BalanceDF to a comma-separated values (csv) file.

        Uses :func:`pd.DataFrame.to_csv`.

        If an 'index' argument is not provided then it defaults to False.

        Args:
            self: Object.
            path_or_buf (Optional[FilePathOrBuffer], optional): location where to save the csv.

        Returns:
            Optional[str]: If path_or_buf is None, returns the resulting csv format as a string. Otherwise returns None.
        """
        if "index" not in kwargs:
            kwargs["index"] = False
        return self.df.to_csv(path_or_buf=path_or_buf, **kwargs)

    ################################################################################
    #  Private API
    ################################################################################

    ##################
    # Column accessors
    ##################
    def _special_columns_names(self: "Sample") -> List[str]:
        """
        Returns names of all special columns (id column,
        wegiht column and outcome columns) in Sample.

        Returns:
            List[str]: names of special columns
        """
        return [
            i.name for i in [self.id_column, self.weight_column] if i is not None
        ] + (
            self._outcome_columns.columns.tolist()
            if self._outcome_columns is not None
            else []
        )

    def _special_columns(self: "Sample") -> pd.DataFrame:
        """
        Returns dataframe of all special columns (id column,
        weight column and outcome columns) in Sample.

        Returns:
            pd.DataFrame: special columns
        """
        return self._df[self._special_columns_names()]

    def _covar_columns_names(self: "Sample") -> List[str]:
        """
        Returns names of all covars in Sample.

        Returns:
            List[str]: names of covars
        """
        return [
            c for c in self._df.columns.values if c not in self._special_columns_names()
        ]

    def _covar_columns(self: "Sample") -> pd.DataFrame:
        """
        Returns dataframe of all covars columns in Sample.

        Returns:
            pd.DataFrame: covars columns
        """
        return self._df[self._covar_columns_names()]

    ################
    #  Errors checks
    ################
    def _check_if_adjusted(self) -> None:
        """
        Raises a ValueError if sample is not adjusted
        """
        if not self.is_adjusted():
            raise ValueError(
                "This is not an adjusted Sample. Use sample.adjust to adjust the sample to target"
            )

    def _no_target_error(self: "Sample") -> None:
        """
        Raises a ValueError if sample doesn't have target
        """
        if not self.has_target():
            raise ValueError(
                "This Sample does not have a target set. Use sample.set_target to add target"
            )

    def _check_outcomes_exists(self) -> None:
        """
        Raises a ValueError if sample doesn't have outcome_columns specified.
        """
        if self.outcomes() is None:
            raise ValueError("This Sample does not have outcome columns specified")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import balance.testutil

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandas.testing

from balance.sample_class import Sample
from balance.stats_and_plots import weighted_comparisons_plots, weighted_stats


s = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3),
            "b": (-42, 8, 2),
            "c": ("x", "y", "z"),
            "id": (1, 2, 3),
            "w": (0.5, 2, 1),
        }
    ),
    id_column="id",
    weight_column="w",
)

s2 = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3),
            "b": (17, 9, -3),
            "c": ("x", "y", "z"),
            "id": (1, 2, 3),
            "w": (0.5, 2, 1),
        }
    ),
    id_column="id",
    weight_column="w",
)


# TODO: split out the weighted_stats.relative_frequency_table function.
class Test_weighted_comparisons_plots(
    balance.testutil.BalanceTestCase,
):
    def test_plot__return_sample_palette(self):
        from balance.stats_and_plots.weighted_comparisons_plots import (
            _return_sample_palette,
        )

        self.assertEqual(
            _return_sample_palette(["self", "target"]),
            {"self": "#de2d26cc", "target": "#9ecae1cc"},
        )
        self.assertEqual(
            _return_sample_palette(["self", "unadjusted"]),
            {"self": "#34a53080", "unadjusted": "#de2d26cc"},
        )
        self.assertEqual(
            _return_sample_palette(["self", "unadjusted", "target"]),
            {"self": "#34a53080", "unadjusted": "#de2d26cc", "target": "#9ecae1cc"},
        )
        self.assertEqual(
            _return_sample_palette(["adjusted", "unadjusted", "target"]),
            {"adjusted": "#34a53080", "unadjusted": "#de2d26cc", "target": "#9ecae1cc"},
        )
        self.assertEqual(_return_sample_palette(["cat", "dog"]), "muted")

    def test_plot__plotly_marker_color(self):
        from balance.stats_and_plots.weighted_comparisons_plots import (
            _plotly_marker_color,
        )

        self.assertEqual(
            _plotly_marker_color("self", True, "color"), "rgba(222,45,38,0.8)"
        )
        self.assertEqual(
            _plotly_marker_color("self", False, "color"), "rgba(52,165,48,0.5)"
        )
        self.assertEqual(
            _plotly_marker_color("self", True, "line"), "rgba(222,45,38,1)"
        )
        self.assertEqual(
            _plotly_marker_color("self", False, "line"), "rgba(52,165,48,1)"
        )

        self.assertEqual(
            _plotly_marker_color("target", True, "color"), "rgb(158,202,225,0.8)"
        )
        self.assertEqual(
            _plotly_marker_color("target", False, "color"), "rgb(158,202,225,0.8)"
        )
        self.assertEqual(
            _plotly_marker_color("target", True, "line"), "rgb(158,202,225,1)"
        )
        self.assertEqual(
            _plotly_marker_color("target", False, "line"), "rgb(158,202,225,1)"
        )

        self.assertEqual(
            _plotly_marker_color("adjusted", False, "color"), "rgba(52,165,48,0.5)"
        )
        self.assertEqual(
            _plotly_marker_color("adjusted", False, "line"), "rgba(52,165,48,1)"
        )

    def test_plot_bar(self):
        import matplotlib.pyplot as plt
        from balance.stats_and_plots.weighted_comparisons_plots import plot_bar

        df = pd.DataFrame(
            {
                "group": ("a", "b", "c", "c"),
                "v1": (1, 2, 3, 4),
            }
        )

        plt.figure(1)
        fig, ax = plt.subplots(1, 1, figsize=(7.2, 7.2))

        plot_bar(
            [
                {"df": df, "weight": pd.Series((1, 1, 1, 1))},
                {"df": df, "weight": pd.Series((2, 1, 1, 1))},
            ],
            names=["self", "target"],
            column="group",
            axis=ax,
            weighted=True,
        )

        self.assertIn("barplot of covar ", ax.get_title())

    def test_plot_dist_weighted_kde_error(self):
        df4 = {
            "df": pd.DataFrame(
                {"numeric_5": pd.Series([0, 0, 0, 0, 0, 1, 1, 2, 3, 4])}
            ),
            "weight": pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 5]),
        }

        plot_dist_type = type(
            weighted_comparisons_plots.plot_dist(
                (df4,),
                dist_type="kde",
                numeric_n_values_threshold=0,
                weighted=False,
                library="seaborn",
                return_axes=True,
            )[0]
        )

        # NOTE: There is no AxesSubplot class until one is invoked and created on the fly.
        #     See details here: https://stackoverflow.com/a/11690800/256662
        self.assertTrue(issubclass(plot_dist_type, plt.Axes))

    def test_relative_frequency_table(self):
        df = pd.DataFrame({"a": list("abcd"), "b": list("bbcd")})

        e = pd.DataFrame({"a": list("abcd"), "prop": (0.25,) * 4})
        self.assertEqual(weighted_stats.relative_frequency_table(df, "a"), e)

        t = weighted_stats.relative_frequency_table(df, "a", pd.Series((1, 1, 1, 1)))
        self.assertEqual(t, e)

        e = pd.DataFrame({"b": list("bcd"), "prop": (0.5, 0.25, 0.25)})
        t = weighted_stats.relative_frequency_table(df, "b")
        self.assertEqual(t, e)

        t = weighted_stats.relative_frequency_table(df, "b", pd.Series((1, 1, 1, 1)))
        self.assertEqual(t, e)

        with self.assertRaisesRegex(TypeError, "must be a pandas Series"):
            weighted_stats.relative_frequency_table(df, "a", 1)

        e = pd.DataFrame({"a": list("abcd"), "prop": (0.2, 0.4, 0.2, 0.2)})
        t = weighted_stats.relative_frequency_table(df, "a", pd.Series((1, 2, 1, 1)))
        self.assertEqual(t, e)

        e = pd.DataFrame({"b": list("bcd"), "prop": (1 / 3, 1 / 3, 1 / 3)})
        t = weighted_stats.relative_frequency_table(
            df, "b", pd.Series((0.5, 0.5, 1, 1))
        )
        self.assertEqual(t, e)

        df = pd.DataFrame(
            {
                "group": ("a", "b", "c", "c"),
                "v1": (1, 2, 3, 4),
            }
        )
        t = weighted_stats.relative_frequency_table(
            df,
            "group",
            pd.Series((2, 1, 1, 1)),
        ).to_dict()
        e = {"group": {0: "a", 1: "b", 2: "c"}, "prop": {0: 0.4, 1: 0.2, 2: 0.4}}
        self.assertEqual(t, e)
        # check that using a pd.Series will give the same results
        t_series = weighted_stats.relative_frequency_table(
            df=df["group"],
            w=pd.Series((2, 1, 1, 1)),
        ).to_dict()
        self.assertEqual(t_series, e)

    def test_naming_plot(self):
        self.assertEqual(
            weighted_comparisons_plots.naming_legend(
                "self", ["self", "target", "unadjusted"]
            ),
            "adjusted",
        )
        self.assertEqual(
            weighted_comparisons_plots.naming_legend(
                "unadjusted", ["self", "target", "unadjusted"]
            ),
            "sample",
        )
        self.assertEqual(
            weighted_comparisons_plots.naming_legend("self", ["self", "target"]),
            "sample",
        )
        self.assertEqual(
            weighted_comparisons_plots.naming_legend("other_name", ["self", "target"]),
            "other_name",
        )

    def test_seaborn_plot_dist(self):
        df4 = {
            "df": pd.DataFrame(
                {"numeric_5": pd.Series([0, 0, 0, 0, 0, 1, 1, 2, 3, 4])}
            ),
            "weight": pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 5]),
        }

        # Check that we get a list of matplotlib axes
        out = []
        for dist_type in ("hist", "kde", "qq", "ecdf"):
            out.append(
                type(
                    weighted_comparisons_plots.seaborn_plot_dist(
                        (df4,),
                        names=["test"],
                        dist_type=dist_type,
                        return_axes=True,
                    )[0]
                )
            )

        # NOTE: There is no AxesSubplot class until one is invoked and created on the fly.
        #     See details here: https://stackoverflow.com/a/11690800/256662
        self.assertTrue(issubclass(out[0], matplotlib.axes.Axes))
        self.assertTrue(issubclass(out[1], matplotlib.axes.Axes))
        self.assertTrue(issubclass(out[2], matplotlib.axes.Axes))
        self.assertTrue(issubclass(out[3], matplotlib.axes.Axes))

    def test_plot_dist(self):

        import plotly.graph_objs as go
        from balance.stats_and_plots.weighted_comparisons_plots import plot_dist
        from numpy import random

        random.seed(96483)

        df = pd.DataFrame(
            {
                "v1": random.random_integers(11111, 11114, size=100).astype(str),
                "v2": random.normal(size=100),
                "v3": random.uniform(size=100),
            }
        ).sort_values(by=["v2"])

        dfs1 = [
            {"df": df, "weight": pd.Series(np.ones(100))},
            {"df": df, "weight": pd.Series(np.ones(99).tolist() + [1000])},
            {"df": df, "weight": pd.Series(np.ones(100))},
        ]

        # If plot_it=False and
        self.assertTrue(
            plot_dist(
                dfs1,
                names=["self", "unadjusted", "target"],
                library="plotly",
                plot_it=False,
            )
            is None
        )

        # check the dict of figures returned:
        dict_of_figures = plot_dist(
            dfs1,
            names=["self", "unadjusted", "target"],
            library="plotly",
            plot_it=False,
            return_dict_of_figures=True,
        )

        self.assertEqual(type(dict_of_figures), dict)
        self.assertEqual(sorted(dict_of_figures.keys()), ["v1", "v2", "v3"])
        self.assertEqual(type(dict_of_figures["v1"]), go.Figure)

        with self.assertRaisesRegex(ValueError, "library must be either*"):
            plot_dist(
                dfs1,
                names=["self", "unadjusted", "target"],
                library="ploting_library_which_is_not_plotly_or_seaborn",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import tempfile

from copy import deepcopy

import balance.testutil
import IPython.display

import numpy as np
import pandas as pd

from balance.sample_class import Sample

# TODO: move s3 and other definitions of sample outsie of classes (from example from TestSamplePrivateAPI)

s1 = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3, 1),
            "b": (-42, 8, 2, -42),
            "o": (7, 8, 9, 10),
            "c": ("x", "y", "z", "v"),
            "id": (1, 2, 3, 4),
            "w": (0.5, 2, 1, 1),
        }
    ),
    id_column="id",
    weight_column="w",
    outcome_columns="o",
)

s2 = Sample.from_frame(
    pd.DataFrame(
        {
            "a": (1, 2, 3),
            "b": (4, 6, 8),
            "id": (1, 2, 3),
            "w": (0.5, 1, 2),
            "c": ("x", "y", "z"),
        }
    ),
    id_column="id",
    weight_column="w",
)

s4 = Sample.from_frame(
    pd.DataFrame(
        {"a": (0, None, 2), "b": (0, None, 2), "c": ("a", "b", "c"), "id": (1, 2, 3)}
    ),
    outcome_columns=("b", "c"),
)


class TestSample(
    balance.testutil.BalanceTestCase,
):
    def test_constructor_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            s1 = Sample()
            print(s1)

    def test_Sample__str__(self):
        self.assertTrue("4 observations x 3 variables" in s1.__str__())
        self.assertTrue("outcome_columns: o" in s1.__str__())
        self.assertTrue("weight_column: w" in s1.__str__())

        self.assertTrue("outcome_columns: None" in s2.__str__())
        self.assertTrue("weight_column: w" in s2.__str__())

        s3 = s1.set_target(s2)
        self.assertTrue("Sample object with target set" in s3.__str__())
        self.assertTrue("target:" in s3.__str__())
        self.assertTrue("3 common variables" in s3.__str__())

        s4 = s3.adjust(method="null")
        self.assertTrue(
            "Adjusted balance Sample object with target set using" in s4.__str__()
        )

    def test_Sample__str__multiple_outcomes(self):
        s1 = Sample.from_frame(
            pd.DataFrame(
                {"a": (1, 2, 3), "b": (4, 6, 8), "id": (1, 2, 3), "w": (0.5, 1, 2)}
            ),
            id_column="id",
            weight_column="w",
            outcome_columns=("a", "b"),
        )
        self.assertTrue("outcome_columns: a,b" in s1.__str__())

    def test_Sample_from_frame(self):
        # test id_column
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)})
        self.assertWarnsRegexp(
            "Guessed id column name id for the data", Sample.from_frame, df
        )
        # TODO: add tests for the two other warnings:
        # self.assertWarnsRegexp("Casting id column to string", Sample.from_frame, df)
        # self.assertWarnsRegexp("No weights passed, setting all weights to 1", Sample.from_frame, df)
        # Using the above would fail since the warnings are sent sequentially and using self.assertWarnsRegexp
        # only catches the first warning.
        self.assertEqual(
            Sample.from_frame(df).id_column, pd.Series((1, 2), name="id").astype(str)
        )
        df = pd.DataFrame({"b": (1, 2), "a": (1, 2)})
        self.assertEqual(
            Sample.from_frame(df, id_column="b").id_column,
            pd.Series((1, 2), name="b").astype(str),
        )
        with self.assertRaisesRegex(
            ValueError,
            "Cannot guess id column name for this DataFrame. Please provide a value in id_column",
        ):
            Sample.from_frame(df)
        with self.assertRaisesRegex(
            ValueError,
            "Dataframe does not have column*",
        ):
            Sample.from_frame(df, id_column="c")

        # test exception if values in id are null
        df = pd.DataFrame({"id": (1, None), "a": (1, 2)})
        with self.assertRaisesRegex(
            ValueError,
            "Null values are not allowed in the id_column",
        ):
            Sample.from_frame(df)

        # test check_id_uniqueness argument
        df = pd.DataFrame({"id": (1, 2, 2)})
        with self.assertRaisesRegex(
            ValueError,
            "Values in the id_column must be unique",
        ):
            Sample.from_frame(df)

        df = pd.DataFrame({"id": (1, 2, 2)})
        self.assertEqual(
            Sample.from_frame(df, check_id_uniqueness=False).df.id,
            pd.Series(("1", "2", "2"), name="id"),
        )

        # test weights_column
        df = pd.DataFrame({"id": (1, 2), "weight": (1, 2)})
        self.assertWarnsRegexp("Guessing weight", Sample.from_frame, df)
        # NOTE how weight that was integer was changed into floats.
        self.assertEqual(
            Sample.from_frame(df).weight_column, pd.Series((1.0, 2.0), name="weight")
        )

        df = pd.DataFrame({"id": (1, 2)})
        self.assertWarnsRegexp("No weights passed", Sample.from_frame, df)
        # NOTE that the default weights are integers, not floats
        # TODO: decide if it's o.k. to keep the default weights be 1s, or change the default to floats
        self.assertEqual(
            Sample.from_frame(df).weight_column, pd.Series((1, 1), name="weight")
        )

        # Test type conversion
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)})
        self.assertEqual(df.a.dtype.type, np.int64)
        self.assertEqual(Sample.from_frame(df).df.a.dtype.type, np.float64)
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)}, dtype=np.int32)
        self.assertEqual(df.a.dtype.type, np.int32)
        self.assertEqual(Sample.from_frame(df).df.a.dtype.type, np.float32)
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)}, dtype=np.int16)
        self.assertEqual(df.a.dtype.type, np.int16)
        self.assertEqual(Sample.from_frame(df).df.a.dtype.type, np.float16)
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)}, dtype=np.int8)
        self.assertEqual(df.a.dtype.type, np.int8)
        self.assertEqual(Sample.from_frame(df).df.a.dtype.type, np.float16)
        # TODO: add tests for other types of conversions

        # Test use_deepcopy
        # after we invoked Sample.from_frame with use_deepcopy=False, we expact the dtype of id to be np.object_
        #   in BOTH the df inside sample and the ORIGINAL df:
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)})
        self.assertEqual(df.id.dtype.type, np.int64)
        self.assertEqual(
            Sample.from_frame(df, use_deepcopy=False).df.id.dtype.type, np.object_
        )
        self.assertEqual(df.id.dtype.type, np.object_)
        # after we invoked Sample.from_frame with use_deepcopy=True (default), we expact the dtype of id to be object_
        #   in the df inside sample, but id in the ORIGINAL df to remain int64:
        df = pd.DataFrame({"id": (1, 2), "a": (1, 2)})
        self.assertEqual(df.id.dtype.type, np.int64)
        self.assertEqual(Sample.from_frame(df).df.id.dtype.type, np.object_)
        self.assertEqual(df.id.dtype.type, np.int64)

    def test_Sample_adjust(self):
        from balance.weighting_methods.adjust_null import adjust_null

        s3 = s1.set_target(s2).adjust(method="null")
        self.assertTrue(s3.is_adjusted())

        s3 = s1.set_target(s2).adjust(method=adjust_null)
        self.assertTrue(s3.is_adjusted())

        # test exception
        with self.assertRaisesRegex(
            ValueError,
            "Method should be one of existing weighting methods",
        ):
            s1.set_target(s2).adjust(method=None)


class TestSample_base_and_adjust_methods(
    balance.testutil.BalanceTestCase,
):
    def test_Sample_df(self):
        # NOTE how integers were changed into floats.
        e = pd.DataFrame(
            {
                "a": (1.0, 2.0, 3.0, 1.0),
                "b": (-42.0, 8.0, 2.0, -42.0),
                "o": (7.0, 8.0, 9.0, 10.0),
                "c": ("x", "y", "z", "v"),
                "id": ("1", "2", "3", "4"),
                "w": (0.5, 2, 1, 1),
            },
            columns=("id", "a", "b", "c", "o", "w"),
        )
        # Verify we get the expected output:
        self.assertEqual(s1.df, e)

        # Check that @property works:
        self.assertTrue(isinstance(Sample.df, property))
        self.assertEqual(Sample.df.fget(s1), s1.df)
        # We can no longer call .df() as if it was a function:
        with self.assertRaisesRegex(TypeError, "'DataFrame' object is not callable"):
            s1.df()

        # NOTE how integers were changed into floats.
        e = pd.DataFrame(
            {
                "a": (1.0, 2.0, 3.0),
                "b": (4.0, 6.0, 8.0),
                "id": ("1", "2", "3"),
                "w": (0.5, 1, 2),
                "c": ("x", "y", "z"),
            },
            columns=("id", "a", "b", "c", "w"),
        )
        self.assertEqual(s2.df, e)

    def test_Sample_outcomes(self):
        # NOTE how integers were changed into floats.
        # TODO: consider removing this test, since it's already tested in test_balancedf.py
        e = pd.DataFrame(
            {
                "o": (7.0, 8.0, 9.0, 10.0),
            },
            columns=["o"],
        )
        self.assertEqual(s1.outcomes().df, e)

    def test_Sample_weights(self):
        e = pd.DataFrame(
            {
                "w": (0.5, 2, 1, 1),
            },
            columns=["w"],
        )
        self.assertEqual(s1.weights().df, e)

    # TODO: consider removing this test, since it's already tested in test_balancedf.py
    def test_Sample_covars(self):
        # NOTE how integers were changed into floats.
        e = pd.DataFrame(
            {
                "a": (1.0, 2.0, 3.0, 1.0),
                "b": (-42.0, 8.0, 2.0, -42.0),
                "c": ("x", "y", "z", "v"),
            }
        )
        self.assertEqual(s1.covars().df, e)

    def test_Sample_model(self):
        np.random.seed(112358)
        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        s = Sample.from_frame(d)

        d = pd.DataFrame(np.random.rand(10000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        t = Sample.from_frame(d)

        a = s.adjust(t, max_de=None, method="null")
        m = a.model()
        self.assertEqual(m["method"], "null_adjustment")
        a = s.adjust(t, max_de=None)
        m = a.model()
        self.assertEqual(m["method"], "ipw")
        # Just test the structure of ipw output
        self.assertTrue("perf" in m.keys())
        self.assertTrue("fit" in m.keys())
        self.assertTrue("coefs" in m["perf"].keys())

    def test_Sample_model_matrix(self):
        #  Main tests for model_matrix are in test_util.py
        s = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": (0, 1, 2),
                    "b": (0, None, 2),
                    "c": ("a", "b", "a"),
                    "id": (1, 2, 3),
                }
            ),
            id_column="id",
        )
        e = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0),
                "b": (0.0, 0.0, 2.0),
                "_is_na_b[T.True]": (0.0, 1.0, 0.0),
                "c[a]": (1.0, 0.0, 1.0),
                "c[b]": (0.0, 1.0, 0.0),
            }
        )
        r = s.model_matrix()
        self.assertEqual(r, e, lazy=True)

    def test_Sample_set_weights(self):
        s = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": (1, 2, 3, 1),
                    "id": (1, 2, 3, 4),
                    "w": (0.5, 2, 1, 1),
                }
            ),
            id_column="id",
            weight_column="w",
        )
        # NOTE that if using set_weights with integers, the weights remain integers
        s.set_weights(pd.Series([1, 2, 3, 4]))
        self.assertEqual(s.weight_column, pd.Series([1, 2, 3, 4], name="w"))
        s.set_weights(pd.Series([1, 2, 3, 4], index=(1, 2, 5, 6)))
        self.assertEqual(
            s.weight_column, pd.Series([np.nan, 1.0, 2.0, np.nan], name="w")
        )
        # test warning
        self.assertWarnsRegexp(
            """Note that not all Sample units will be assigned weights""",
            Sample.set_weights,
            s,
            pd.Series([1, 2, 3, 4], index=(1, 2, 5, 6)),
        )
        # no warning
        self.assertNotWarns(
            Sample.set_weights,
            s,
            pd.Series([1, 2, 3, 4], index=(0, 1, 2, 3)),
        )

    def test_Sample_set_unadjusted(self):
        s5 = s1.set_unadjusted(s2)
        self.assertTrue(s5._links["unadjusted"] is s2)
        # test exceptions when there is no asecond sample
        with self.assertRaisesRegex(
            TypeError,
            "set_unadjusted must be called with second_sample argument of type Sample",
        ):
            s1.set_unadjusted("Not a Sample object")

    def test_Sample_is_adjusted(self):
        self.assertFalse(s1.is_adjusted())
        # TODO: move definitions of s3 outside of function
        s3 = s1.set_target(s2)
        self.assertFalse(s3.is_adjusted())
        # TODO: move definitions of s3 outside of function
        s3_adjusted = s3.adjust(method="null")
        self.assertTrue(s3_adjusted.is_adjusted())

    def test_Sample_set_target(self):
        s5 = s1.set_target(s2)
        self.assertTrue(s5._links["target"] is s2)
        # test exceptions when the provided object is not a second sample
        with self.assertRaisesRegex(
            ValueError,
            "A target, a Sample object, must be specified",
        ):
            s1.set_target("Not a Sample object")

    def test_Sample_has_target(self):
        self.assertFalse(s1.has_target())
        self.assertTrue(s1.set_target(s2).has_target())


class TestSample_metrics_methods(
    balance.testutil.BalanceTestCase,
):
    def test_Sample_covar_means(self):
        # TODO: take definition of s3_null outside of function
        s3_null = s1.adjust(s2, method="null")
        e = pd.DataFrame(
            {
                "a": [(0.5 * 1 + 2 * 2 + 3 * 1 + 1 * 1) / (0.5 + 2 + 1 + 1)],
                "b": [(-42 * 0.5 + 8 * 2 + 2 * 1 + -42 * 1) / (0.5 + 2 + 1 + 1)],
                "c[x]": [(1 * 0.5) / (0.5 + 2 + 1 + 1)],
                "c[y]": [(1 * 2) / (0.5 + 2 + 1 + 1)],
                "c[z]": [(1 * 1) / (0.5 + 2 + 1 + 1)],
                "c[v]": [(1 * 1) / (0.5 + 2 + 1 + 1)],
            }
        ).transpose()
        e = pd.concat((e,) * 2, axis=1, sort=True)
        e = pd.concat(
            (
                e,
                pd.DataFrame(
                    {
                        "a": [(1 * 0.5 + 2 * 1 + 3 * 2) / (0.5 + 1 + 2)],
                        "b": [(4 * 0.5 + 6 * 1 + 8 * 2) / (0.5 + 1 + 2)],
                        "c[x]": [(1 * 0.5) / (0.5 + 1 + 2)],
                        "c[y]": [(1 * 1) / (0.5 + 1 + 2)],
                        "c[z]": [(1 * 2) / (0.5 + 1 + 2)],
                        "c[v]": np.nan,
                    }
                ).transpose(),
            ),
            axis=1,
            sort=True,
        )
        e.columns = pd.Series(("unadjusted", "adjusted", "target"), name="source")
        self.assertEqual(s3_null.covar_means(), e)

        # test exceptions when there is no adjusted
        with self.assertRaisesRegex(
            ValueError,
            "This is not an adjusted Sample. Use sample.adjust to adjust the sample to target",
        ):
            s1.covar_means()

    def test_Sample_design_effect(self):
        self.assertEqual(s1.design_effect().round(3), 1.235)
        self.assertEqual(s4.design_effect(), 1.0)

    def test_Sample_design_effect_prop(self):
        # TODO: take definition of s3_null outside of function
        s3_null = s1.adjust(s2, method="null")
        self.assertEqual(s3_null.design_effect_prop(), 0.0)

        # tetst exceptions when there is no adjusted
        with self.assertRaisesRegex(
            ValueError,
            "This is not an adjusted Sample. Use sample.adjust to adjust the sample to target",
        ):
            s1.design_effect_prop()

    def test_Sample_outcome_sd_prop(self):
        # TODO: take definition of s3_null outside of function
        s3_null = s1.adjust(s2, method="null")
        self.assertEqual(s3_null.outcome_sd_prop(), pd.Series((0.0), index=["o"]))
        # test with two outcomes
        # TODO: take definition of s1_two_outcomes outside of function
        s1_two_outcomes = Sample.from_frame(
            pd.DataFrame(
                {
                    "o1": (7, 8, 9, 10),
                    "o2": (7, 8, 9, 11),
                    "c": ("x", "y", "z", "y"),
                    "id": (1, 2, 3, 4),
                    "w": (0.5, 2, 1, 1),
                },
            ),
            id_column="id",
            weight_column="w",
            outcome_columns=["o1", "o2"],
        )
        s3_null = s1_two_outcomes.adjust(s2, method="null")
        self.assertEqual(
            s3_null.outcome_sd_prop(), pd.Series((0.0, 0.0), index=["o1", "o2"])
        )

        # tetst exceptions when there is no adjusted
        with self.assertRaisesRegex(
            ValueError,
            "This Sample does not have outcome columns specified",
        ):
            s2.adjust(s2, method="null").outcome_sd_prop()

    def test_outcome_variance_ratio(self):
        from balance.stats_and_plots.weighted_stats import weighted_var

        # Testing it also works with outcomes
        np.random.seed(112358)

        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        t = Sample.from_frame(d)

        d = pd.DataFrame(np.random.rand(1000, 11))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghijk"[i] for i in range(0, 11)})
        d["b"] = np.sqrt(d["b"])

        a_with_outcome = Sample.from_frame(d, outcome_columns=["k"])
        a_with_outcome_adjusted = a_with_outcome.adjust(t, max_de=1.5)

        # verifying this does what we expact it does:
        self.assertEqual(
            round(a_with_outcome_adjusted.outcome_variance_ratio()[0], 5),
            round(
                (
                    weighted_var(
                        a_with_outcome_adjusted.outcomes().df,
                        a_with_outcome_adjusted.weights().df["weight"],
                    )
                    / weighted_var(
                        a_with_outcome_adjusted._links["unadjusted"].outcomes().df,
                        a_with_outcome_adjusted._links["unadjusted"]
                        .weights()
                        .df["weight"],
                    )
                )[0],
                5,
            ),
        )

        self.assertEqual(
            round(a_with_outcome_adjusted.outcome_variance_ratio()[0], 5), 0.97516
        )

        # two outcomes, with no adjustment (var ratio should be 1)
        a_with_outcome = Sample.from_frame(d, outcome_columns=["j", "k"])
        a_with_outcome_adjusted = a_with_outcome.adjust(t, method="null")
        self.assertEqual(
            a_with_outcome_adjusted.outcome_variance_ratio(),
            pd.Series([1.0, 1.0], index=["j", "k"]),
        )

    def test_Sample_weights_summary(self):
        self.assertEqual(
            s1.weights().summary().to_dict(),
            {"self": {"w": 1.125}, "self_ci": {"w": (0.591, 1.659)}},
        )
        self.assertEqual(s1.weights().summary().iloc[0, 0], (0.5 + 2 + 1 + 1) / 4)

    def test_Sample_summary(self):
        s1_summ = s1.summary()
        self.assertTrue("Model performance" not in s1_summ)
        self.assertTrue("Covar ASMD" not in s1_summ)

        s3 = s1.set_target(s2)
        s3_summ = s3.summary()
        self.assertTrue("Model performance" not in s1_summ)
        self.assertTrue("Covar ASMD (6 variables)" in s3_summ)
        self.assertTrue("design effect" not in s3_summ)

        s3 = s3.set_unadjusted(s1)
        s3_summ = s3.summary()
        self.assertTrue("Covar ASMD reduction: 0.0%" in s3_summ)
        self.assertTrue("Covar ASMD (6 variables)" in s3_summ)
        self.assertTrue("->" in s3_summ)
        self.assertTrue("design effect" in s3_summ)

        s3 = s1.set_target(s2).adjust(method="null")
        s3_summ = s3.summary()
        self.assertTrue("Covar ASMD reduction: 0.0%" in s3_summ)
        self.assertTrue("design effect" in s3_summ)

    def test_Sample_invalid_outcomes(self):
        with self.assertRaisesRegex(
            ValueError,
            r"outcome columns \['o'\] not in df columns \['a', 'id', 'weight'\]",
        ):
            Sample.from_frame(
                pd.DataFrame({"a": (1, 2, 3, 1), "id": (1, 2, 3, 4)}),
                outcome_columns="o",
            )

    def test_Sample_diagnostics(self):
        import numpy as np
        import pandas as pd

        # TODO (p2): move the objects created here outside of this function and possible make this simpler.
        from balance.sample_class import Sample

        np.random.seed(112358)

        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        s = Sample.from_frame(d)

        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        t = Sample.from_frame(d)

        a = s.adjust(t)
        a_diagnostics = a.diagnostics()
        # print(a_diagnostics)

        self.assertEqual(a_diagnostics.shape, (198, 3))
        self.assertEqual(a_diagnostics.columns.to_list(), ["metric", "val", "var"])
        self.assertEqual(
            a_diagnostics[a_diagnostics["metric"] == "adjustment_method"]["var"].values,
            np.array(["ipw"]),
        )

        output = a_diagnostics.groupby("metric").size().to_dict()
        expected = {
            "adjustment_failure": 1,
            "covar_asmd_adjusted": 11,
            "covar_asmd_improvement": 11,
            "covar_asmd_unadjusted": 11,
            "covar_main_asmd_adjusted": 11,
            "covar_main_asmd_improvement": 11,
            "covar_main_asmd_unadjusted": 11,
            "model_coef": 92,
            "model_glance": 10,
            "adjustment_method": 1,
            "size": 4,
            "weights_diagnostics": 24,
        }
        self.assertEqual(output, expected)

        b = s.adjust(t, method="cbps")
        b_diagnostics = b.diagnostics()
        # print(b_diagnostics)

        self.assertEqual(b_diagnostics.shape, (196, 3))
        self.assertEqual(b_diagnostics.columns.to_list(), ["metric", "val", "var"])
        self.assertEqual(
            b_diagnostics[b_diagnostics["metric"] == "adjustment_method"]["var"].values,
            np.array(["cbps"]),
        )

        output = b_diagnostics.groupby("metric").size().to_dict()
        expected = {
            "adjustment_failure": 1,
            "balance_optimize_result": 2,
            "gmm_optimize_result_bal_init": 2,
            "gmm_optimize_result_glm_init": 2,
            "rescale_initial_result": 2,
            "beta_optimal": 92,
            "covar_asmd_adjusted": 11,
            "covar_asmd_improvement": 11,
            "covar_asmd_unadjusted": 11,
            "covar_main_asmd_adjusted": 11,
            "covar_main_asmd_improvement": 11,
            "covar_main_asmd_unadjusted": 11,
            "adjustment_method": 1,
            "size": 4,
            "weights_diagnostics": 24,
        }
        self.assertEqual(output, expected)

        c = s.adjust(t, method="null")
        c_diagnostics = c.diagnostics()
        self.assertEqual(c_diagnostics.shape, (96, 3))
        self.assertEqual(c_diagnostics.columns.to_list(), ["metric", "val", "var"])
        self.assertEqual(
            c_diagnostics[c_diagnostics["metric"] == "adjustment_method"]["var"].values,
            np.array(["null_adjustment"]),
        )

    def test_Sample_keep_only_some_rows_columns(self):
        import numpy as np
        import pandas as pd

        # TODO (p2): move the objects created here outside of this function and possible make this simpler.
        from balance.sample_class import Sample

        np.random.seed(112358)

        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        d["b"] = np.sqrt(d["b"])
        s = Sample.from_frame(d)

        d = pd.DataFrame(np.random.rand(1000, 10))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})
        t = Sample.from_frame(d)

        a = s.adjust(t, max_de=1.5)

        # if both rows_to_keep = None, columns_to_keep = None - then keep_only_some_rows_columns returns the same object
        self.assertTrue(
            a is a.keep_only_some_rows_columns(rows_to_keep=None, columns_to_keep=None)
        )

        # let's remove some columns and rows:
        a2 = a.keep_only_some_rows_columns(
            rows_to_keep=None, columns_to_keep=["b", "c"]
        )

        # Making sure asmd works
        output_orig = a.covars().asmd().round(2).to_dict()
        output_new = a2.covars().asmd().round(2).to_dict()
        expected_orig = {
            "j": {"self": 0.01, "unadjusted": 0.03, "unadjusted - self": 0.02},
            "i": {"self": 0.02, "unadjusted": 0.0, "unadjusted - self": -0.02},
            "h": {"self": 0.04, "unadjusted": 0.09, "unadjusted - self": 0.04},
            "g": {"self": 0.0, "unadjusted": 0.0, "unadjusted - self": 0.0},
            "f": {"self": 0.01, "unadjusted": 0.03, "unadjusted - self": 0.02},
            "e": {"self": 0.0, "unadjusted": 0.0, "unadjusted - self": 0.0},
            "d": {"self": 0.05, "unadjusted": 0.12, "unadjusted - self": 0.06},
            "c": {"self": 0.04, "unadjusted": 0.05, "unadjusted - self": 0.01},
            "b": {"self": 0.14, "unadjusted": 0.55, "unadjusted - self": 0.41},
            "a": {"self": 0.01, "unadjusted": 0.0, "unadjusted - self": -0.01},
            "mean(asmd)": {"self": 0.03, "unadjusted": 0.09, "unadjusted - self": 0.05},
        }
        expected_new = {
            "c": {"self": 0.04, "unadjusted": 0.05, "unadjusted - self": 0.01},
            "b": {"self": 0.14, "unadjusted": 0.55, "unadjusted - self": 0.41},
            "mean(asmd)": {"self": 0.09, "unadjusted": 0.3, "unadjusted - self": 0.21},
        }

        self.assertEqual(output_orig, expected_orig)
        self.assertEqual(output_new, expected_new)

        # Making sure diagnostics works, and also seeing we got change in
        # what we expect
        a_diag = a.diagnostics()
        a2_diag = a2.diagnostics()
        a_diag_tbl = a_diag.groupby("metric").size().to_dict()
        a2_diag_tbl = a2_diag.groupby("metric").size().to_dict()

        # The mean weight should be 1 (since we normalize for the sum of weights to be equal to len(weights))
        ss = a_diag.eval("(metric == 'weights_diagnostics') & (var == 'describe_mean')")
        self.assertEqual(round(float(a_diag[ss].val), 4), 1.000)

        # keeping only columns 'b' and 'c' leads to have only 3 asmd instead of 11:
        self.assertEqual(a_diag_tbl["covar_main_asmd_adjusted"], 11)
        self.assertEqual(a2_diag_tbl["covar_main_asmd_adjusted"], 3)

        # now we get only 2 covars counted instead of 10:
        ss_condition = "(metric == 'size') & (var == 'sample_covars')"
        ss = a_diag.eval(ss_condition)
        ss2 = a2_diag.eval(ss_condition)
        self.assertEqual(int(a_diag[ss].val), 10)
        self.assertEqual(int(a2_diag[ss2].val), 2)

        # And the mean asmd is different
        ss_condition = "(metric == 'covar_main_asmd_adjusted') & (var == 'mean(asmd)')"
        ss = a_diag.eval(ss_condition)
        ss2 = a2_diag.eval(ss_condition)
        self.assertEqual(round(float(a_diag[ss].val), 4), 0.0338)
        self.assertEqual(round(float(a2_diag[ss2].val), 3), 0.093)

        # Also checking filtering using rows_to_keep:
        a3 = a.keep_only_some_rows_columns(
            rows_to_keep="a>0.5", columns_to_keep=["b", "c"]
        )

        # Making sure the weights are of the same length as the df
        self.assertEqual(a3.df.shape[0], a3.weights().df.shape[0])

        # Making sure asmd works - we can see it's different then for a2
        output_new = a3.covars().asmd().round(2).to_dict()
        expected_new = {
            "c": {"self": 0.06, "unadjusted": 0.07, "unadjusted - self": 0.01},
            "b": {"self": 0.21, "unadjusted": 0.61, "unadjusted - self": 0.4},
            "mean(asmd)": {"self": 0.13, "unadjusted": 0.34, "unadjusted - self": 0.21},
        }
        self.assertEqual(output_new, expected_new)

        a3_diag = a3.diagnostics()
        a3_diag_tbl = a3_diag.groupby("metric").size().to_dict()

        # The structure of the diagnostics table is the same with and without
        # the filtering. So when comparing a3 to a2, we should get the same results:
        # i.e.: a2_diag_tbl == a3_diag_tbl # True
        self.assertEqual(a2_diag_tbl, a3_diag_tbl)
        # However, the number of samples is different!
        ss_condition = "(metric == 'size') & (var == 'sample_obs')"
        self.assertEqual(int(a_diag[a_diag.eval(ss_condition)].val), 1000)
        self.assertEqual(int(a2_diag[a2_diag.eval(ss_condition)].val), 1000)
        self.assertEqual(int(a3_diag[a3_diag.eval(ss_condition)].val), 508)
        # also in the target
        ss_condition = "(metric == 'size') & (var == 'target_obs')"
        self.assertEqual(int(a_diag[a_diag.eval(ss_condition)].val), 1000)
        self.assertEqual(int(a2_diag[a2_diag.eval(ss_condition)].val), 1000)
        self.assertEqual(
            int(a3_diag[a3_diag.eval(ss_condition)].val), 516
        )  # since a<0.5 is different for target!
        # also in the weights
        ss = a_diag.eval(
            "(metric == 'weights_diagnostics') & (var == 'describe_count')"
        )
        self.assertEqual(int(a_diag[ss].val), 1000)
        ss = a3_diag.eval(
            "(metric == 'weights_diagnostics') & (var == 'describe_count')"
        )
        self.assertEqual(int(a3_diag[ss].val), 508)
        # Notice also that the calculated values from the weights are different
        ss = a_diag.eval("(metric == 'weights_diagnostics') & (var == 'design_effect')")
        self.assertEqual(round(float(a_diag[ss].val), 4), 1.493)
        ss = a3_diag.eval(
            "(metric == 'weights_diagnostics') & (var == 'design_effect')"
        )
        self.assertEqual(round(float(a3_diag[ss].val), 4), 1.4802)

        # Testing it also works with outcomes
        np.random.seed(112358)

        d = pd.DataFrame(np.random.rand(1000, 11))
        d["id"] = range(0, d.shape[0])
        d = d.rename(columns={i: "abcdefghijk"[i] for i in range(0, 11)})
        d["b"] = np.sqrt(d["b"])
        a_with_outcome = Sample.from_frame(d, outcome_columns=["k"])
        a_with_outcome_adjusted = a_with_outcome.adjust(t, max_de=1.5)

        # We can also filter using an outcome variable (althought this would NOT filter on target)
        # a proper logger warning is issued
        self.assertEqual(
            a_with_outcome_adjusted.keep_only_some_rows_columns(
                rows_to_keep="k>0.5"
            ).df.shape,
            (481, 13),
        )

        a_with_outcome_adjusted2 = a_with_outcome_adjusted.keep_only_some_rows_columns(
            rows_to_keep="b>0.5", columns_to_keep=["b", "c"]
        )

        self.assertEqual(
            a_with_outcome_adjusted2.outcomes().mean().round(3).to_dict(),
            {"k": {"self": 0.491, "unadjusted": 0.494}},
        )

        # TODO (p2): possibly add checks for columns_to_keep = None while doing something with rows_to_keep

        # test if only some columns exsists
        self.assertWarnsRegexp(
            "Note that not all columns_to_keep are in Sample",
            s1.keep_only_some_rows_columns,
            columns_to_keep=["g", "a"],
        )
        self.assertEqual(
            s1.keep_only_some_rows_columns(
                columns_to_keep=["g", "a"]
            )._df.columns.tolist(),
            ["a"],
        )


class TestSample_to_download(balance.testutil.BalanceTestCase):
    def test_Sample_to_download(self):
        r = s1.to_download()
        self.assertIsInstance(r, IPython.display.FileLink)

    def test_Sample_to_csv(self):
        with tempfile.NamedTemporaryFile() as tf:
            s1.to_csv(path_or_buf=tf.name)
            r = tf.read()
            e = (
                b"id,a,b,c,o,w\n1,1,-42,x,7,0.5\n"
                b"2,2,8,y,8,2.0\n3,3,2,z,9,1.0\n4,1,-42,v,10,1.0\n"
            )
            self.assertTrue(r, e)


class TestSamplePrivateAPI(balance.testutil.BalanceTestCase):
    def test__links(self):
        self.assertTrue(len(s1._links.keys()) == 0)

        s3 = s1.set_target(s2)
        self.assertTrue(s3._links["target"] is s2)
        self.assertTrue(s3.has_target())

        s3_adjusted = s3.adjust(method="null")
        self.assertTrue(s3_adjusted._links["target"] is s2)
        self.assertTrue(s3_adjusted._links["unadjusted"] is s3)
        self.assertTrue(s3_adjusted.has_target())

    def test__special_columns_names(self):
        self.assertEqual(
            sorted(s4._special_columns_names()), ["b", "c", "id", "weight"]
        )

    # NOTE how integers were changed into floats.
    def test__special_columns(self):
        # NOTE how integers in weight were changed into floats.
        self.assertEqual(
            s4._special_columns(),
            pd.DataFrame(
                {
                    "id": ("1", "2", "3"),
                    # Weights were filled automatically to be integers of 1s:
                    "weight": (1, 1, 1),
                    "b": (0.0, None, 2.0),
                    "c": ("a", "b", "c"),
                }
            ),
        )

    def test__covar_columns_names(self):
        self.assertEqual(sorted(s1._covar_columns_names()), ["a", "b", "c"])

    def test__covar_columns(self):
        # NOTE how integers were changed into floats.
        self.assertEqual(
            s1._covar_columns(),
            pd.DataFrame(
                {
                    "a": (1.0, 2.0, 3.0, 1.0),
                    "b": (-42.0, 8.0, 2.0, -42.0),
                    "c": ("x", "y", "z", "v"),
                }
            ),
        )

    def test_Sample__check_if_adjusted(self):
        with self.assertRaisesRegex(
            ValueError,
            "This is not an adjusted Sample. Use sample.adjust to adjust the sample to target",
        ):
            s1._check_if_adjusted()
        # TODO: move definitions of s3 outside of function
        s3 = s1.set_target(s2)
        with self.assertRaisesRegex(
            ValueError,
            "This is not an adjusted Sample. Use sample.adjust to adjust the sample to target",
        ):
            s3._check_if_adjusted()
        # TODO: move definitions of s3 outside of function
        s3_adjusted = s3.adjust(method="null")
        self.assertTrue(
            s3_adjusted._check_if_adjusted() is None
        )  # Does not raise an error

    def test_Sample__no_target_error(self):
        # test exception when the is no taret
        with self.assertRaisesRegex(
            ValueError,
            "This Sample does not have a target set. Use sample.set_target to add target",
        ):
            s1._no_target_error()
        s3 = s1.set_target(s2)
        s3._no_target_error()  # Should not raise an error

    def test_Sample__check_outcomes_exists(self):
        with self.assertRaisesRegex(
            ValueError,
            "This Sample does not have outcome columns specified",
        ):
            s2._check_outcomes_exists()
        self.assertTrue(s1._check_outcomes_exists() is None)  # Does not raise an error


class TestSample_NA_behavior(balance.testutil.BalanceTestCase):
    def test_can_handle_various_NAs(self):
        # Testing if we can handle NA values from pandas

        def get_sample_to_adjust(df, standardize_types=True):
            s1 = Sample.from_frame(df, standardize_types=standardize_types)
            s2 = deepcopy(s1)
            s2.set_weights(np.ones(100))
            return s1.set_target(s2)

        np.random.seed(123)
        df = pd.DataFrame(
            {
                "a": np.random.uniform(size=100),
                "c": np.random.choice(
                    ["a", "b", "c", "d"],
                    size=100,
                    replace=True,
                    p=[0.01, 0.04, 0.5, 0.45],
                ),
                "id": range(100),
                "weight": np.random.uniform(size=100) + 0.5,
            }
        )

        # This works fine
        smpl_to_adj = get_sample_to_adjust(df)
        self.assertIsInstance(smpl_to_adj.adjust(method="ipw"), Sample)

        # This should raise a TypeError:
        with self.assertRaisesRegex(
            TypeError,
            "boolean value of NA is ambiguous",
        ):
            smpl_to_adj = get_sample_to_adjust(df)
            # smpl_to_adj._df.iloc[0, 0] = pd.NA
            smpl_to_adj._df.iloc[0, 1] = pd.NA
            # This will raise the error:
            smpl_to_adj.adjust(method="ipw")

        # This should raise a TypeError:
        with self.assertRaisesRegex(
            Exception,
            "series must be numeric",
        ):
            # Adding NA to a numeric column turns it into an object.
            # This raises an error in util.quantize
            smpl_to_adj = get_sample_to_adjust(df)
            smpl_to_adj._df.iloc[0, 0] = pd.NA
            # smpl_to_adj._df.iloc[0, 1] = pd.NA
            # This will raise the error:
            smpl_to_adj.adjust(method="ipw")

        # This works fine
        df.iloc[0, 0] = np.nan
        df.iloc[0, 1] = np.nan
        smpl_to_adj = get_sample_to_adjust(df)
        self.assertIsInstance(smpl_to_adj.adjust(method="ipw"), Sample)

        # This also works fine (thanks to standardize_types=True)
        df.iloc[0, 0] = pd.NA
        df.iloc[0, 1] = pd.NA
        smpl_to_adj = get_sample_to_adjust(df)
        self.assertIsInstance(smpl_to_adj.adjust(method="ipw"), Sample)

        # Turning standardize_types to False should raise a TypeError (since we have pd.NA):
        with self.assertRaisesRegex(
            TypeError,
            "boolean value of NA is ambiguous",
        ):
            # df.iloc[0, 0] = pd.NA
            df.iloc[0, 1] = pd.NA
            smpl_to_adj = get_sample_to_adjust(df, standardize_types=False)
            smpl_to_adj.adjust(method="ipw")
